from __future__ import annotations

from abc import abstractmethod
from typing import Iterable, Optional, TypedDict, Union

import numpy as np
from PIL import Image

from turftopic.data import TopicData

UrlStr = str

ImageRepr = Union[Image.Image, UrlStr, None]


def _load_images(images: Iterable[ImageRepr]) -> Iterable[Image]:
    for image in images:
        if image is None:
            yield None
        elif isinstance(image, str):
            yield Image.open(image)
        else:
            yield image


def _naive_join_embeddings(
    text_embeddings: np.ndarray, image_embeddings: np.ndarray
) -> np.ndarray:
    """Produces document embeddings by averaging text and image embeddings"""
    return np.nanmean(np.stack([text_embeddings, image_embeddings]), axis=0)


class MultimodalEmbeddings(TypedDict):
    """Dict type for embeddings in a multimodal context.

    Exclusively used for type checking."""

    text_embeddings: np.ndarray
    image_embeddings: np.ndarray
    document_embeddings: np.ndarray


def encode_multimodal(
    encoder, sentences: list[str], images: list[ImageRepr]
) -> dict[str, np.ndarray]:
    """Produce multimodal embeddings of the documents passed to the model.

    Parameters
    ----------
    encoder
        MTEB or SentenceTransformer compatible embedding model.
    sentences: list[str]
        Textual documents to encode.
    images: list[ImageRepr]
        Corresponding images for each document.

    Returns
    -------
    MultimodalEmbeddings
        Text, image and joint document embeddings.
    """
    if len(sentences) != len(images):
        raise ValueError("Images and documents were not the same length.")
    if hasattr(encoder, "get_text_embeddings"):
        text_embeddings = np.array(encoder.get_text_embeddings(sentences))
    else:
        text_embeddings = encoder.encode(sentences)
    embedding_size = text_embeddings.shape[1]
    images = list(_load_images(images))
    if hasattr(encoder, "get_image_embeddings"):
        image_embeddings = np.array(encoder.get_image_embeddings(images))
    else:
        image_embeddings = []
        for image in images:
            if image is not None:
                image_embeddings.append(encoder.encode(image))
            else:
                image_embeddings.append(np.full(embedding_size, np.nan))
        image_embeddings = np.stack(image_embeddings)
    if hasattr(encoder, "get_fused_embeddings"):
        document_embeddings = np.array(
            encoder.get_fused_embeddings(
                texts=sentences,
                images=images,
            )
        )
    else:
        document_embeddings = _naive_join_embeddings(
            text_embeddings, image_embeddings
        )
    return {
        "text_embeddings": text_embeddings,
        "image_embeddings": image_embeddings,
        "document_embeddings": document_embeddings,
    }


class MultimodalModel:
    """Base model for multimodal topic models."""

    def encode_multimodal(
        self,
        sentences: list[str],
        images: list[ImageRepr],
    ) -> dict[str, np.ndarray]:
        """Produce multimodal embeddings of the documents passed to the model.

        Parameters
        ----------
        sentences: list[str]
            Textual documents to encode.
        images: list[ImageRepr]
            Corresponding images for each document.

        Returns
        -------
        MultimodalEmbeddings
            Text, image and joint document embeddings.

        """
        return encode_multimodal(self.encoder_, sentences, images)

    @staticmethod
    def validate_embeddings(embeddings: Optional[MultimodalEmbeddings]):
        if embeddings is None:
            return
        try:
            document_embeddings = embeddings["document_embeddings"]
            image_embeddings = embeddings["image_embeddings"]
        except KeyError as e:
            raise TypeError(
                "embeddings do not contain document and image embeddings, can't be used for multimodal modelling."
            ) from e
        if document_embeddings.shape != image_embeddings.shape:
            raise ValueError(
                f"Shape mismatch between document_embeddings {document_embeddings.shape} and image_embeddings {image_embeddings.shape}"
            )

    def validate_encoder(self):
        if not hasattr(self.encoder_, "encode"):
            if not all(
                (
                    hasattr(self.encoder_, "get_text_embeddings"),
                    hasattr(self.encoder_, "get_image_embeddings"),
                ),
            ):
                raise TypeError(
                    "An encoder must either have an encode() method or a get_text_embeddings and get_image_embeddings method (optionally get_fused_embeddings)"
                )

    @abstractmethod
    def fit_transform_multimodal(
        self,
        raw_documents: list[str],
        images: list[ImageRepr],
        y=None,
        embeddings: Optional[MultimodalEmbeddings] = None,
    ) -> np.ndarray:
        """Fits topic model in a multimodal context and returns the document-topic matrix.

        Parameters
        ----------
        raw_documents: iterable of str
            Documents to fit the model on.
        images: list[ImageRepr]
            Images corresponding to each document.
        y: None
            Ignored, exists for sklearn compatibility.
        embeddings: MultimodalEmbeddings
            Precomputed multimodal embeddings.

        Returns
        -------
        ndarray of shape (n_documents, n_topics)
            Document-topic matrix.
        """
        pass

    def fit_multimodal(
        self,
        raw_documents: list[str],
        images: list[ImageRepr],
        y=None,
        embeddings: Optional[MultimodalEmbeddings] = None,
    ):
        """Fits topic model on a multimodal corpus.

        Parameters
        ----------
        raw_documents: iterable of str
            Documents to fit the model on.
        images: list[ImageRepr]
            Images corresponding to each document.
        y: None
            Ignored, exists for sklearn compatibility.
        embeddings: MultimodalEmbeddings
            Precomputed multimodal embeddings.

        Returns
        -------
        Self
            The fitted topic model
        """
        self.fit_transform_multimodal(raw_documents, images, y, embeddings)
        return self

    @staticmethod
    def collect_top_images(
        images: list[Image.Image],
        image_topic_matrix: np.ndarray,
        n_images: int = 20,
        negative: bool = False,
    ) -> list[list[Image.Image]]:
        top_images: list[list[Image.Image]] = []
        for image_topic_vector in image_topic_matrix.T:
            if negative:
                image_topic_vector = -image_topic_vector
            top_im_ind = np.argsort(-image_topic_vector)[:20]
            top_im = [images[int(i)] for i in top_im_ind]
            top_images.append(top_im)
        return top_images

    def prepare_multimodal_topic_data(
        self,
        corpus: list[str],
        images: list[ImageRepr],
        embeddings: Optional[MultimodalEmbeddings] = None,
    ) -> TopicData:
        """Produces multimodal topic inference data for a given corpus, that can be then used and reused.
        Exists to allow visualizations out of the box with topicwizard.

        Parameters
        ----------
        corpus: list[str]
            Documents to infer topical content for.
        images: list[ImageRepr]
            Images belonging to the documents.
        embeddings: MultimodalEmbeddings
            Embeddings of documents.

        Returns
        -------
        TopicData
            Information about topical inference in a dictionary.
        """
        if embeddings is None:
            embeddings = self.encode_multimodal(corpus, images)
        document_topic_matrix = self.fit_transform_multimodal(
            corpus, images=images, embeddings=embeddings
        )
        dtm = self.vectorizer.transform(corpus)  # type: ignore
        try:
            classes = self.classes_
        except AttributeError:
            classes = list(range(self.components_.shape[0]))
        res = TopicData(
            corpus=corpus,
            document_term_matrix=dtm,
            vocab=self.get_vocab(),
            document_topic_matrix=document_topic_matrix,
            document_representation=embeddings["document_embeddings"],
            topic_term_matrix=self.components_,  # type: ignore
            transform=getattr(self, "transform", None),
            topic_names=self.topic_names,
            classes=classes,
            has_negative_side=self.has_negative_side,
            hierarchy=getattr(self, "hierarchy", None),
            images=images,
            top_images=self.top_images,
            negative_images=getattr(self, "negative_images", None),
        )
        return res



import math
import random
from typing import Optional, Union

import numpy as np
import pyro
import pyro.distributions as dist
import torch
import torch.nn as nn
import torch.nn.functional as F
from pyro.infer import SVI, TraceMeanField_ELBO
from rich.console import Console
from sentence_transformers import SentenceTransformer
from sklearn.feature_extraction.text import CountVectorizer

from turftopic.base import ContextualModel, Encoder
from turftopic.encoders.multimodal import MultimodalEncoder
from turftopic.multimodal import (
    Image,
    ImageRepr,
    MultimodalEmbeddings,
    MultimodalModel,
)
from turftopic.vectorizers.default import default_vectorizer


class EncoderNetwork(nn.Module):
    def __init__(self, contextualized_size, num_topics, hidden, dropout):
        super().__init__()
        self.drop = nn.Dropout(dropout)  # to avoid component collapse
        self.fc1 = nn.Linear(contextualized_size, hidden)
        self.fc2 = nn.Linear(hidden, hidden)
        self.fcmu = nn.Linear(hidden, num_topics)
        self.fclv = nn.Linear(hidden, num_topics)
        self.bnmu = nn.BatchNorm1d(num_topics, affine=False)
        self.bnlv = nn.BatchNorm1d(num_topics, affine=False)

    def forward(self, inputs):
        h = F.softplus(self.fc1(inputs))
        h = F.softplus(self.fc2(h))
        h = self.drop(h)
        logtheta_loc = self.bnmu(self.fcmu(h))
        logtheta_logvar = self.bnlv(self.fclv(h))
        logtheta_scale = (0.5 * logtheta_logvar).exp()  # Enforces positivity
        return logtheta_loc, logtheta_scale


class DecoderNetwork(nn.Module):
    def __init__(self, vocab_size, num_topics, dropout):
        super().__init__()
        self.beta = nn.Linear(num_topics, vocab_size, bias=False)
        self.bn = nn.BatchNorm1d(vocab_size, affine=False)
        self.drop = nn.Dropout(dropout)

    def forward(self, inputs):
        inputs = self.drop(inputs)
        return F.softmax(self.bn(self.beta(inputs)), dim=1)


class Model(nn.Module):
    def __init__(
        self, vocab_size, contextualized_size, num_topics, hidden, dropout
    ):
        super().__init__()
        self.vocab_size = vocab_size
        self.num_topics = num_topics
        self.encoder = EncoderNetwork(
            contextualized_size, num_topics, hidden, dropout
        )
        self.decoder = DecoderNetwork(vocab_size, num_topics, dropout)

    def model(self, bow, contextualized):
        pyro.module("decoder", self.decoder)
        with pyro.plate("documents", bow.shape[0]):
            logtheta_loc = bow.new_zeros((bow.shape[0], self.num_topics))
            logtheta_scale = bow.new_ones((bow.shape[0], self.num_topics))
            logtheta = pyro.sample(
                "logtheta",
                dist.Normal(logtheta_loc, logtheta_scale).to_event(1),
            )
            theta = F.softmax(logtheta, -1)
            count_param = self.decoder(theta)
            total_count = int(bow.sum(-1).max())
            pyro.sample(
                "obs", dist.Multinomial(total_count, count_param), obs=bow
            )

    def guide(self, bow, contextualized):
        pyro.module("encoder", self.encoder)
        with pyro.plate("documents", contextualized.shape[0]):
            logtheta_loc, logtheta_scale = self.encoder(contextualized)
            logtheta = pyro.sample(
                "logtheta",
                dist.Normal(logtheta_loc, logtheta_scale).to_event(1),
            )

    def beta(self):
        return self.decoder.beta.weight.cpu().detach().T


class AutoEncodingTopicModel(ContextualModel, MultimodalModel):
    """Variational autoencoding topic models
    with contextualized representations (CTM).
    Uses amortized variational inference with neural networks
    to estimate posterior for ProdLDA.

    ```python
    from turftopic import AutoEncodingTopicModel

    corpus: list[str] = ["some text", "more text", ...]

    model = AutoEncodingTopicModel(10, combined=False).fit(corpus)
    model.print_topics()
    ```

    Parameters
    ----------
    n_components: int
        Number of topics.
    encoder: str or SentenceTransformer
        Model to encode documents/terms, all-MiniLM-L6-v2 is the default.
    vectorizer: CountVectorizer, default None
        Vectorizer used for term extraction.
        Can be used to prune or filter the vocabulary.
    combined: bool, default False
        Indicates whether encoder inputs should be combined
        with bow representations.
        When False the model is equivalent to ZeroShotTM,
        when True it is CombinedTM.
    dropout_rate: float, default 0.1
        Dropout in the encoder layers.
    hidden: int, default 100
        Size of hidden layers in the encoder network.
    batch_size: int, default 42
        Batch size when training the network.
    learning_rate: float, default 1e-2
        Learning rate for the optimizer.
    n_epochs: int, default 50
        Number of epochs to run during training.
    random_state: int, default None
        Random state to use so that results are exactly reproducible.
    """

    def __init__(
        self,
        n_components: int,
        encoder: Union[
            Encoder, SentenceTransformer, MultimodalEncoder
        ] = "sentence-transformers/all-MiniLM-L6-v2",
        vectorizer: Optional[CountVectorizer] = None,
        combined: bool = False,
        dropout_rate: float = 0.1,
        hidden: int = 100,
        batch_size: int = 42,
        learning_rate: float = 1e-2,
        n_epochs: int = 50,
        random_state: Optional[int] = None,
    ):
        self.n_components = n_components
        self.random_state = random_state
        self.encoder = encoder
        if isinstance(encoder, str):
            self.encoder_ = SentenceTransformer(encoder)
        else:
            self.encoder_ = encoder
        self.validate_encoder()
        if vectorizer is None:
            self.vectorizer = default_vectorizer()
        else:
            self.vectorizer = vectorizer
        self.combined = combined
        self.dropout_rate = dropout_rate
        self.batch_size = batch_size
        self.n_epochs = n_epochs
        self.learning_rate = learning_rate
        self.hidden = hidden

    def transform(
        self, raw_documents, embeddings: Optional[np.ndarray] = None
    ):
        """Infers topic importances for new documents based on a fitted model.

        Parameters
        ----------
        raw_documents: iterable of str
            Documents to fit the model on.
        embeddings: ndarray of shape (n_documents, n_dimensions), optional
            Precomputed document encodings.

        Returns
        -------
        ndarray of shape (n_dimensions, n_topics)
            Document-topic matrix.
        """
        if embeddings is None:
            embeddings = self.encoder_.encode(raw_documents)
        if self.combined:
            bow = self.vectorizer.fit_transform(raw_documents)
            contextual_embeddings = np.concatenate(
                (embeddings, bow.toarray()), axis=1
            )
        else:
            contextual_embeddings = embeddings
        contextual_embeddings = torch.tensor(contextual_embeddings).float()
        loc, scale = self.model.encoder(contextual_embeddings)
        prob = torch.softmax(loc, dim=-1)
        return prob.cpu().data.numpy()

    def fit(
        self, raw_documents, y=None, embeddings: Optional[np.ndarray] = None
    ):
        self.fit_transform(raw_documents, y, embeddings=embeddings)
        return self

    def fit_transform(
        self, raw_documents, y=None, embeddings: Optional[np.ndarray] = None
    ) -> np.ndarray:
        console = Console()
        with console.status("Fitting model") as status:
            if embeddings is None:
                status.update("Encoding documents")
                embeddings = self.encoder_.encode(raw_documents)
                console.log("Documents encoded.")
            status.update("Extracting terms.")
            document_term_matrix = self.vectorizer.fit_transform(raw_documents)
            console.log("Term extraction done.")
            seed = self.random_state or random.randint(0, 10_000)
            torch.manual_seed(seed)
            pyro.set_rng_seed(seed)
            device = torch.device("cpu")
            pyro.clear_param_store()
            contextualized_size = embeddings.shape[1]
            if self.combined:
                contextualized_size = (
                    contextualized_size + document_term_matrix.shape[1]
                )
            self.model = Model(
                vocab_size=document_term_matrix.shape[1],
                contextualized_size=contextualized_size,
                num_topics=self.n_components,
                hidden=self.hidden,
                dropout=self.dropout_rate,
            )
            self.model.to(device)
            optimizer = pyro.optim.Adam({"lr": self.learning_rate})
            svi = SVI(
                self.model.model,
                self.model.guide,
                optimizer,
                loss=TraceMeanField_ELBO(),
            )
            num_batches = int(
                math.ceil(document_term_matrix.shape[0] / self.batch_size)
            )

            status.update(f"Fitting model. Epoch [0/{self.n_epochs}]")
            for epoch in range(self.n_epochs):
                running_loss = 0.0
                for i in range(num_batches):
                    batch_bow = np.atleast_2d(
                        document_term_matrix[
                            i * self.batch_size : (i + 1) * self.batch_size, :
                        ].toarray()
                    )
                    # Skipping batches that are smaller than 2
                    if batch_bow.shape[0] < 2:
                        continue
                    batch_contextualized = np.atleast_2d(
                        embeddings[
                            i * self.batch_size : (i + 1) * self.batch_size, :
                        ]
                    )
                    if self.combined:
                        batch_contextualized = np.concatenate(
                            (batch_contextualized, batch_bow), axis=1
                        )
                    batch_contextualized = (
                        torch.tensor(batch_contextualized).float().to(device)
                    )
                    batch_bow = torch.tensor(batch_bow).float().to(device)
                    loss = svi.step(batch_bow, batch_contextualized)
                    running_loss += loss / batch_bow.size(0)
                status.update(
                    f"Fitting model. Epoch [{epoch}/{self.n_epochs}], Loss [{running_loss}]"
                )
            self.components_ = np.array(self.model.beta())
            console.log("Model fitting done.")
            document_topic_matrix = self.transform(
                raw_documents, embeddings=embeddings
            )
            self.top_documents = self.get_top_documents(
                raw_documents, document_topic_matrix=document_topic_matrix
            )
            return document_topic_matrix

    def fit_transform_multimodal(
        self,
        raw_documents: list[str],
        images: list[ImageRepr],
        y=None,
        embeddings: Optional[MultimodalEmbeddings] = None,
    ) -> np.ndarray:
        self.validate_embeddings(embeddings)
        console = Console()
        self.multimodal_embeddings = embeddings
        with console.status("Fitting model") as status:
            if self.multimodal_embeddings is None:
                status.update("Encoding documents")
                self.multimodal_embeddings = self.encode_multimodal(
                    raw_documents, images
                )
                console.log("Documents encoded.")
            status.update("Extracting terms.")
            document_term_matrix = self.vectorizer.fit_transform(raw_documents)
            console.log("Term extraction done.")
            seed = self.random_state or random.randint(0, 10_000)
            torch.manual_seed(seed)
            pyro.set_rng_seed(seed)
            device = torch.device("cpu")
            pyro.clear_param_store()
            contextualized_size = self.multimodal_embeddings[
                "document_embeddings"
            ].shape[1]
            if self.combined:
                contextualized_size = (
                    contextualized_size + document_term_matrix.shape[1]
                )
            self.model = Model(
                vocab_size=document_term_matrix.shape[1],
                contextualized_size=contextualized_size,
                num_topics=self.n_components,
                hidden=self.hidden,
                dropout=self.dropout_rate,
            )
            self.model.to(device)
            optimizer = pyro.optim.Adam({"lr": self.learning_rate})
            svi = SVI(
                self.model.model,
                self.model.guide,
                optimizer,
                loss=TraceMeanField_ELBO(),
            )
            num_batches = int(
                math.ceil(document_term_matrix.shape[0] / self.batch_size)
            )

            status.update(f"Fitting model. Epoch [0/{self.n_epochs}]")
            for epoch in range(self.n_epochs):
                running_loss = 0.0
                for i in range(num_batches):
                    batch_bow = np.atleast_2d(
                        document_term_matrix[
                            i * self.batch_size : (i + 1) * self.batch_size, :
                        ].toarray()
                    )
                    # Skipping batches that are smaller than 2
                    if batch_bow.shape[0] < 2:
                        continue
                    batch_contextualized = np.atleast_2d(
                        self.multimodal_embeddings["document_embeddings"][
                            i * self.batch_size : (i + 1) * self.batch_size, :
                        ]
                    )
                    if self.combined:
                        batch_contextualized = np.concatenate(
                            (batch_contextualized, batch_bow), axis=1
                        )
                    batch_contextualized = (
                        torch.tensor(batch_contextualized).float().to(device)
                    )
                    batch_bow = torch.tensor(batch_bow).float().to(device)
                    loss = svi.step(batch_bow, batch_contextualized)
                    running_loss += loss / batch_bow.size(0)
                status.update(
                    f"Fitting model. Epoch [{epoch}/{self.n_epochs}], Loss [{running_loss}]"
                )
            self.components_ = np.array(self.model.beta())
            console.log("Model fitting done.")
            status.update("Transforming documents and images.")
            document_topic_matrix = self.transform(
                raw_documents,
                embeddings=self.multimodal_embeddings["document_embeddings"],
            )
            self.image_topic_matrix = self.transform(
                raw_documents,
                embeddings=self.multimodal_embeddings["image_embeddings"],
            )
            self.top_images: list[list[Image.Image]] = self.collect_top_images(
                images, self.image_topic_matrix
            )
            self.top_documents = self.get_top_documents(
                raw_documents, document_topic_matrix=document_topic_matrix
            )
            console.log("Transformation done.")
        return document_topic_matrix



from datetime import datetime
from functools import partial
from typing import Literal, Optional, Union

import numpy as np
from rich.console import Console
from sentence_transformers import SentenceTransformer
from sklearn.exceptions import NotFittedError
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.manifold import TSNE
from sklearn.metrics.pairwise import cosine_similarity

from turftopic._datamapplot import build_datamapplot
from turftopic.base import ContextualModel, Encoder
from turftopic.dynamic import DynamicTopicModel
from turftopic.encoders.multimodal import MultimodalEncoder
from turftopic.models._snmf import SNMF
from turftopic.multimodal import (
    ImageRepr,
    MultimodalEmbeddings,
    MultimodalModel,
)
from turftopic.optimization import (
    optimize_n_components,
)
from turftopic.vectorizers.default import default_vectorizer

NOT_MATCHING_ERROR = (
    "Document embedding dimensionality ({n_dims}) doesn't match term embedding dimensionality ({n_word_dims}). "
    + "Perhaps you are using precomputed embeddings but forgot to pass an encoder to your model. "
    + "Try to initialize the model with the encoder you used for computing the embeddings."
)


def bic_snmf(
    n_components: int, sparsity: float, X, random_state: int = 42
) -> float:
    decomp = SNMF(
        n_components=n_components,
        sparsity=sparsity,
        random_state=42,
        verbose=False,
        progress_bar=False,
    )
    doc_topic = decomp.fit_transform(X)
    rss = np.square(decomp.reconstruction_err_)
    n_docs, n_dims = X.shape
    # BIC1 from https://pmc.ncbi.nlm.nih.gov/articles/PMC9181460/
    bic1 = np.log(rss) + n_components * (
        (n_docs + n_dims) / (n_docs * n_dims)
    ) * np.log((n_docs * n_dims) / (n_docs + n_dims))
    return bic1


class SensTopic(ContextualModel, DynamicTopicModel, MultimodalModel):
    """Semi-nonnegative Semantic Signal Separation.

    ```python
    from turftopic import SensTopic

    corpus: list[str] = ["some text", "more text", ...]

    model = SensTopic(10).fit(corpus)
    model.print_topics()
    ```

    Parameters
    ----------
    n_components: int, default "auto"
        Number of topics.
        If "auto", the number of topics is determined using BIC.
    encoder: str or SentenceTransformer
        Model to encode documents/terms, all-MiniLM-L6-v2 is the default.
    vectorizer: CountVectorizer, default None
        Vectorizer used for term extraction.
        Can be used to prune or filter the vocabulary.
    max_iter: int, default 200
        Maximum number of iterations for S-NMF.
    feature_importance: "axial", "angular" or "combined", default "combined"
        Defines whether the word's position on an axis ('axial'), it's angle to the axis ('angular')
        or their combination ('combined') should determine the word's importance for a topic.
    random_state: int, default None
        Random state to use so that results are exactly reproducible.
    sparsity: float, default 1
        L1 penalty applied to document-topic proportions.
        Higher values push the model to assign fewer topics to a single document,
        while lower values will distribute topics across documents.
    """

    def __init__(
        self,
        n_components: Union[int, Literal["auto"]] = "auto",
        encoder: Union[
            Encoder, str, MultimodalEncoder
        ] = "sentence-transformers/all-MiniLM-L6-v2",
        vectorizer: Optional[CountVectorizer] = None,
        max_iter: int = 200,
        feature_importance: Literal[
            "axial", "angular", "combined"
        ] = "combined",
        random_state: Optional[int] = None,
        sparsity: float = 1,
    ):
        self.n_components = n_components
        self.encoder = encoder
        self.feature_importance = feature_importance
        if isinstance(encoder, str):
            self.encoder_ = SentenceTransformer(encoder)
        else:
            self.encoder_ = encoder
        self.validate_encoder()
        if vectorizer is None:
            self.vectorizer = default_vectorizer()
        else:
            self.vectorizer = vectorizer
        self.max_iter = max_iter
        self.random_state = random_state
        self.sparsity = sparsity

    def estimate_components(
        self, feature_importance: Literal["axial", "angular", "combined"]
    ) -> np.ndarray:
        """Reestimates components based on the chosen feature_importance method."""
        if feature_importance == "axial":
            self.components_ = self.axial_components_
        elif feature_importance == "angular":
            self.components_ = self.angular_components_
        elif feature_importance == "combined":
            self.components_ = (
                np.square(self.axial_components_) * self.angular_components_
            )
        if hasattr(self, "axial_temporal_components_"):
            if feature_importance == "axial":
                self.temporal_components_ = self.axial_temporal_components_
            elif feature_importance == "angular":
                self.temporal_components_ = self.angular_temporal_components_
            elif feature_importance == "combined":
                self.temporal_components_ = (
                    np.square(self.axial_temporal_components_)
                    * self.angular_temporal_components_
                )
        return self.components_

    def fit_transform(
        self, raw_documents, y=None, embeddings: Optional[np.ndarray] = None
    ) -> np.ndarray:
        console = Console()
        self.embeddings = embeddings
        with console.status("Fitting model") as status:
            if self.embeddings is None:
                status.update("Encoding documents")
                self.embeddings = self.encoder_.encode(raw_documents)
                console.log("Documents encoded.")
            if self.n_components == "auto":
                status.update("Finding the number of components.")
                self.n_components_ = optimize_n_components(
                    partial(
                        bic_snmf, X=self.embeddings, sparsity=self.sparsity
                    ),
                    min_n=1,
                    verbose=True,
                )
                console.log("N components set at: " + str(self.n_components_))
            else:
                self.n_components_ = self.n_components
            self.decomposition = SNMF(
                self.n_components_,
                max_iter=self.max_iter,
                sparsity=self.sparsity,
                random_state=self.random_state,
            )
            status.update("Decomposing embeddings")
            doc_topic = self.decomposition.fit_transform(self.embeddings, y=y)
            console.log("Decomposition done.")
            status.update("Extracting terms.")
            vocab = self.vectorizer.fit(raw_documents).get_feature_names_out()
            console.log("Term extraction done.")
            if getattr(self, "vocab_embeddings", None) is None:
                status.update("Encoding vocabulary")
                self.vocab_embeddings = self.encoder_.encode(vocab)
            if self.vocab_embeddings.shape[1] != self.embeddings.shape[1]:
                raise ValueError(
                    NOT_MATCHING_ERROR.format(
                        n_dims=self.embeddings.shape[1],
                        n_word_dims=self.vocab_embeddings.shape[1],
                    )
                )
            console.log("Vocabulary encoded.")
            status.update("Estimating term importances")
            vocab_topic = self.decomposition.transform(self.vocab_embeddings)
            self.axial_components_ = vocab_topic.T
            if self.feature_importance == "axial":
                self.components_ = self.axial_components_
            elif self.feature_importance == "angular":
                self.components_ = self.angular_components_
            elif self.feature_importance == "combined":
                self.components_ = (
                    np.square(self.axial_components_)
                    * self.angular_components_
                )
            self.top_documents = self.get_top_documents(
                raw_documents, document_topic_matrix=doc_topic
            )
            self.document_topic_matrix = doc_topic
            console.log("Model fitting done.")
        return doc_topic

    def transform(self, raw_documents, embeddings=None):
        if embeddings is None:
            embeddings = self.encoder_.encode(raw_documents)
        return self.decomposition.transform(embeddings)

    def fit_transform_multimodal(
        self,
        raw_documents: list[str],
        images: list[ImageRepr],
        y=None,
        embeddings: Optional[MultimodalEmbeddings] = None,
    ) -> np.ndarray:
        self.validate_embeddings(embeddings)
        console = Console()
        self.images = images
        self.multimodal_embeddings = embeddings
        with console.status("Fitting model") as status:
            if self.multimodal_embeddings is None:
                status.update("Encoding documents")
                self.multimodal_embeddings = self.encode_multimodal(
                    raw_documents, images
                )
                console.log("Documents encoded.")
            self.embeddings = self.multimodal_embeddings["document_embeddings"]
            if self.n_components == "auto":
                status.update("Finding the number of components.")
                self.n_components_ = optimize_n_components(
                    partial(
                        bic_snmf, X=self.embeddings, sparsity=self.sparsity
                    ),
                    min_n=1,
                    verbose=True,
                )
                console.log("N components set at: " + str(self.n_components_))
            else:
                self.n_components_ = self.n_components
            self.decomposition = SNMF(
                self.n_components_,
                max_iter=self.max_iter,
                sparsity=self.sparsity,
                random_state=self.random_state,
            )
            status.update("Decomposing embeddings")
            doc_topic = self.decomposition.fit_transform(self.embeddings, y=y)
            console.log("Decomposition done.")
            status.update("Extracting terms.")
            vocab = self.vectorizer.fit(raw_documents).get_feature_names_out()
            console.log("Term extraction done.")
            status.update("Encoding vocabulary")
            self.vocab_embeddings = self.encode_documents(vocab)
            if self.vocab_embeddings.shape[1] != self.embeddings.shape[1]:
                raise ValueError(
                    NOT_MATCHING_ERROR.format(
                        n_dims=self.embeddings.shape[1],
                        n_word_dims=self.vocab_embeddings.shape[1],
                    )
                )
            console.log("Vocabulary encoded.")
            status.update("Estimating term importances")
            vocab_topic = self.decomposition.transform(self.vocab_embeddings)
            self.axial_components_ = vocab_topic.T
            if self.feature_importance == "axial":
                self.components_ = self.axial_components_
            elif self.feature_importance == "angular":
                self.components_ = self.angular_components_
            elif self.feature_importance == "combined":
                self.components_ = (
                    np.square(self.axial_components_)
                    * self.angular_components_
                )
            console.log("Model fitting done.")
            status.update("Transforming images")
            self.image_topic_matrix = self.transform(
                [], embeddings=self.multimodal_embeddings["image_embeddings"]
            )
            self.top_images = self.collect_top_images(
                images, self.image_topic_matrix
            )
            self.top_documents = self.get_top_documents(
                raw_documents, document_topic_matrix=doc_topic
            )
            console.log("Images transformed")
        return doc_topic

    def fit_transform_dynamic(
        self,
        raw_documents,
        timestamps: list[datetime],
        embeddings: Optional[np.ndarray] = None,
        bins: Union[int, list[datetime]] = 10,
    ) -> np.ndarray:
        document_topic_matrix = self.fit_transform(
            raw_documents, embeddings=embeddings
        )
        time_labels, self.time_bin_edges = self.bin_timestamps(
            timestamps, bins
        )
        n_comp, n_vocab = self.components_.shape
        n_bins = len(self.time_bin_edges) - 1
        self.axial_temporal_components_ = np.full(
            (n_bins, n_comp, n_vocab),
            np.nan,
            dtype=self.components_.dtype,
        )
        self.temporal_importance_ = np.zeros((n_bins, n_comp))
        # doc_topic = np.dot(X, self.components_.T)
        for i_timebin in np.unique(time_labels):
            topic_importances = document_topic_matrix[
                time_labels == i_timebin
            ].mean(axis=0)
            self.temporal_importance_[i_timebin, :] = topic_importances
            t_doc_topic = document_topic_matrix[time_labels == i_timebin]
            t_embeddings = self.embeddings[time_labels == i_timebin]
            t_components = self.decomposition.fit_timeslice(
                t_embeddings, t_doc_topic
            )
            ax_t = np.maximum(
                self.vocab_embeddings @ np.linalg.pinv(t_components), 0
            )
            self.axial_temporal_components_[i_timebin, :, :] = ax_t.T
        self.estimate_components(self.feature_importance)
        return document_topic_matrix

    @property
    def angular_components_(self):
        """Reweights words based on their angle in ICA-space to the axis
        base vectors.
        """
        if not hasattr(self, "axial_components_"):
            raise NotFittedError("Model has not been fitted yet.")
        word_vectors = self.axial_components_.T
        n_topics = self.axial_components_.shape[0]
        axis_vectors = np.eye(n_topics)
        cosine_components = cosine_similarity(axis_vectors, word_vectors)
        return cosine_components

    @property
    def angular_temporal_components_(self):
        """Reweights words based on their angle in ICA-space to the axis
        base vectors in a dynamic model.
        """
        if not hasattr(self, "axial_temporal_components_"):
            raise NotFittedError("Model has not been fitted dynamically.")
        components = []
        for axial_components in self.axial_temporal_components_:
            word_vectors = axial_components.T
            n_topics = axial_components.shape[0]
            axis_vectors = np.eye(n_topics)
            cosine_components = cosine_similarity(axis_vectors, word_vectors)
            components.append(cosine_components)
        return np.stack(components)

    def plot_topic_decay(self):
        try:
            import plotly.graph_objects as go
        except (ImportError, ModuleNotFoundError) as e:
            raise ModuleNotFoundError(
                "Please install plotly if you intend to use plots in Turftopic."
            ) from e
        doc_topic = self.document_topic_matrix
        topic_proportions = []
        for dt in doc_topic:
            sum_dt = dt.sum()
            if sum_dt > 0:
                dt /= sum_dt
            dt = -np.sort(-dt)
            topic_proportions.append(dt)
        topic_proportions = np.stack(topic_proportions)
        med_prop = np.median(topic_proportions, axis=0)
        upper = np.quantile(topic_proportions, 0.975, axis=0)
        lower = np.quantile(topic_proportions, 0.025, axis=0)
        fig = go.Figure(
            [
                go.Scatter(
                    name="Median",
                    x=np.arange(self.n_components_),
                    y=med_prop,
                    mode="lines",
                    line=dict(color="rgb(31, 119, 180)"),
                ),
                go.Scatter(
                    name="Upper Bound",
                    x=np.arange(self.n_components_),
                    y=upper,
                    mode="lines",
                    marker=dict(color="#444"),
                    line=dict(width=0),
                    showlegend=False,
                ),
                go.Scatter(
                    name="Lower Bound",
                    x=np.arange(self.n_components_),
                    y=lower,
                    marker=dict(color="#444"),
                    line=dict(width=0),
                    mode="lines",
                    fillcolor="rgba(68, 68, 68, 0.3)",
                    fill="tonexty",
                    showlegend=False,
                ),
            ]
        )
        fig = fig.update_layout(
            template="plotly_white",
            xaxis_title="Topic Rank",
            yaxis_title="Topic Proportion",
            title="Topic Decay",
            font=dict(family="Merriweather", size=16),
        )
        return fig

    def plot_components(
        self, hover_text: Optional[list[str]] = None, **kwargs
    ):
        """Creates an interactive browser plot of the topics in your data using plotly.

        Parameters
        ----------
        hover_text: list of str, optional
            Text to show when hovering over a document.

        Returns
        -------
        plot
            Interactive datamap plot, you can call the `.show()` method to
            display it in your default browser or save it as static HTML using `.write_html()`.
        """
        doc_topic = self.document_topic_matrix
        coords = TSNE(2, metric="cosine").fit_transform(doc_topic)
        labels = np.argmax(doc_topic, axis=1)
        print(np.unique_counts(labels))
        topics_present = np.sort(np.unique(labels))
        names = [self.topic_names[i] for i in topics_present]
        if getattr(self, "topic_descriptions", None) is not None:
            desc = [self.topic_descriptions[i] for i in topics_present]
        else:
            desc = None
        all_words = self.get_top_words()
        keywords = [all_words[i] for i in topics_present]
        fig = build_datamapplot(
            coords,
            labels=labels,
            topic_names=names,
            top_words=keywords,
            hover_text=hover_text,
            topic_descriptions=desc,
            classes=topics_present,
            # Boundaries are unlikely to be very clear
            cluster_boundary_polygons=False,
        )
        return fig

    def plot_components_datamapplot(
        self, hover_text: Optional[list[str]] = None, **kwargs
    ):
        """Creates an interactive browser plot of the topics in your data using datamapplot.

        Parameters
        ----------
        hover_text: list of str, optional
            Text to show when hovering over a document.

        Returns
        -------
        plot
            Interactive datamap plot, you can call the `.show()` method to
            display it in your default browser or save it as static HTML using `.write_html()`.
        """
        doc_topic = self.document_topic_matrix
        coords = TSNE(2, metric="cosine").fit_transform(doc_topic)
        labels = np.argmax(doc_topic, axis=1)
        print(np.unique_counts(labels))
        topics_present = np.sort(np.unique(labels))
        names = [self.topic_names[i] for i in topics_present]
        if getattr(self, "topic_descriptions", None) is not None:
            desc = [self.topic_descriptions[i] for i in topics_present]
        else:
            desc = None
        all_words = self.get_top_words()
        keywords = [all_words[i] for i in topics_present]
        fig = build_datamapplot(
            coords,
            labels=labels,
            topic_names=names,
            top_words=keywords,
            hover_text=hover_text,
            topic_descriptions=desc,
            classes=topics_present,
            # Boundaries are unlikely to be very clear
            cluster_boundary_polygons=False,
        )
        return fig



from turftopic.analyzers.hf_llm import LLMAnalyzer
from turftopic.analyzers.t5 import T5Analyzer
from turftopic.error import NotInstalled

try:
    from turftopic.analyzers.openai import OpenAIAnalyzer
except ModuleNotFoundError:
    OpenAITopicNamer = NotInstalled("OpenAIAnalyzer", "openai")


__all__ = ["T5Analyzer", "LLMAnalyzer", "OpenAIAnalyzer"]



import csv
import io

import numpy as np
from sklearn.preprocessing import label_binarize


def safe_binarize(y, classes) -> np.ndarray:
    """label_binarize, but its behaviour stays consistent when the labels are {0,1}"""
    if set(classes) == {0, 1}:
        binary_vec = np.squeeze(label_binarize(y, classes=classes))
        negative_vec = np.zeros(binary_vec.shape[0], dtype=binary_vec.dtype)
        negative_vec[binary_vec == 0] = 1
        # Concatenating
        onehot = np.stack((binary_vec, negative_vec))
        return onehot.T
    else:
        return label_binarize(y, classes=classes)


def export_table(
    table: list[list[str]],
    format="csv",
) -> str:
    if format == "csv":
        output = io.StringIO()
        writer = csv.writer(output, quoting=csv.QUOTE_NONNUMERIC)
        for row in table:
            writer.writerow(row)
        return output.getvalue()
    if format == "latex":
        columns, *rows = table
        n_columns = len(columns)
        latex_column_format = " ".join(["l"] * n_columns)
        latex_columns = " & ".join(columns) + "\\\\"
        latex_rows = [" & ".join(row) + "\\\\" for row in rows]
        latex_table = [
            "\\begin{center}",
            f"\\begin{{tabular}}{{ {latex_column_format} }}",
            latex_columns,
            "\\hline",
            *latex_rows,
            "\\end{tabular}",
            "\\end{center}",
        ]
        return "\n".join(latex_table)
    if format == "markdown":
        columns, *rows = table
        n_columns = len(columns)
        separator = ["-"] * n_columns
        markdown_rows = [
            "|" + "|".join([" " + value + " " for value in row]) + "|"
            for row in [columns, separator, *rows]
        ]
        return "\n".join(markdown_rows)
    raise ValueError(
        f"Format '{format}' not supported for tables, please use 'markdown', 'latex' or 'csv'"
    )


def sanitize_for_html(text: str) -> str:
    """Sanitizes strings so they can be put into JS or HTML strings"""
    # Escaping special characters
    text = (
        text.replace("&", "&amp;")
        .replace("<", "&lt;")
        .replace(">", "&gt;")
        .replace('"', "&quot;")
        .replace("'", "&#39;")
    )
    # Removing unnecessary whitespace
    text = " ".join(text.split())
    return text


def confidence_ellipse(mean, cov, n_std=1, size=100):
    pearson = cov[0, 1] / np.sqrt(cov[0, 0] * cov[1, 1])
    ell_radius_x = np.sqrt(1 + pearson)
    ell_radius_y = np.sqrt(1 - pearson)
    theta = np.linspace(0, 2 * np.pi, size)
    ellipse_coords = np.column_stack(
        [ell_radius_x * np.cos(theta), ell_radius_y * np.sin(theta)]
    )
    x_scale = np.sqrt(cov[0, 0]) * n_std
    y_scale = np.sqrt(cov[1, 1]) * n_std
    x_mean, y_mean = mean
    translation_matrix = np.tile(
        [x_mean, y_mean], (ellipse_coords.shape[0], 1)
    )
    rotation_matrix = np.array(
        [
            [np.cos(np.pi / 4), np.sin(np.pi / 4)],
            [-np.sin(np.pi / 4), np.cos(np.pi / 4)],
        ]
    )
    scale_matrix = np.array([[x_scale, 0], [0, y_scale]])
    ellipse_coords = (
        ellipse_coords.dot(rotation_matrix).dot(scale_matrix)
        + translation_matrix
    )
    path = f"M {ellipse_coords[0, 0]}, {ellipse_coords[0, 1]}"
    for k in range(1, len(ellipse_coords)):
        path += f"L{ellipse_coords[k, 0]}, {ellipse_coords[k, 1]}"
    path += " Z"
    return path



import tempfile
import time
import webbrowser
from random import shuffle
from typing import Literal

import numpy as np

from turftopic._figure import HTMLFigure
from turftopic.utils import sanitize_for_html


def open_html(html: str):
    with tempfile.NamedTemporaryFile("w", delete=False, suffix=".html") as f:
        url = "file://" + f.name
        f.write(html)
        time.sleep(1.0)
        webbrowser.open(url)


COLOR_PALETTE = [
    "rgba(116,221,201, 0.25)",
    "rgba(235,171,204, 0.25)",
    "rgba(181,217,160, 0.25)",
    "rgba(197,179,239, 0.25)",
    "rgba(222,202,142, 0.25)",
    "rgba(160,188,233, 0.25)",
    "rgba(234,177,158, 0.25)",
    "rgba(129,209,231, 0.25)",
    "rgba(207,198,216, 0.25)",
    "rgba(186,213,195, 0.25)",
]


def create_bar_plot(seed_id, topic_names, topic_sizes, colors: list[str]):
    topic_sizes = [int(size) for size in topic_sizes]
    plot_js = """
    const yValue{seed_id} = {topic_sizes};
    const data{seed_id} = [
        {{
            x: {topic_names},
            y: yValue{seed_id},
            type: "bar",
            hoverinfo: 'none',
            textposition: 'auto',
            text: yValue{seed_id}.map(n => `n=${{n}}`).map(String),
            marker: {{
                color: {colors},
                line: {{
                color: 'black',
                width: 1
                }}
            }}
        }}
    ];
    const layout{seed_id} = {{
        font: {{family: "Merriweather, sans-serif", size: 16}},
        margin: {{t: 0, l: 30, b:15, r: 30}},
        height: 200,
        width: 600,
        yaxis: {{
            showticklabels: false
        }},
        xaxis: {{
            showticklabels: false
        }},
    }};
    Plotly.newPlot("plot-{seed_id}", data{seed_id}, layout{seed_id})
    """.format(
        topic_names=str(topic_names),
        topic_sizes=str(topic_sizes),
        colors=str(colors),
        seed_id=str(seed_id),
    )
    res = """
    <div id="plot-{seed_id}" class="plot"></div>
    <script>
    {plot_js}
    </script>
    """.format(
        plot_js=plot_js,
        seed_id=seed_id,
    )
    return res


def render_cards(
    seed_id,
    topic_names,
    keywords,
    topic_descriptions,
    topic_sizes,
    colors: list[str],
):
    # Adding the ID if it doesn't already start like that
    topic_names = [
        f"{i} - {name}"
        for i, name in enumerate(topic_names)
        if not name.startswith(str(i))
    ]
    res = create_bar_plot(seed_id, topic_names, topic_sizes, colors)
    for i, (name, keys, desc, color) in enumerate(
        zip(
            topic_names,
            keywords,
            topic_descriptions,
            colors,
        ),
    ):
        desc = sanitize_for_html(desc)
        name = sanitize_for_html(name)
        res += """
        <div class="card" style="background-color: {color};", id="card-{seed_id}-{i}">
        <h3>{name}</h3>
        <p><b>Keywords: </b><i>{keywords}</i></p>
        <p><b>Description: </b>{description}<p>
        </div>
        """.format(
            name=name,
            keywords=", ".join(keys),
            description=desc,
            i=i,
            color=color,
            seed_id=seed_id,
        )
    return res


def prep_document(doc: str, max_chars: int = 900):
    if len(doc) > max_chars:
        doc = doc[: max_chars - 3] + "..."
    doc = sanitize_for_html(doc)
    return doc


def render_documents(top_documents: list[list[str]], colors: list[str]):
    documents = []
    labels = []
    # Flattening the array out
    for topic_id, _top in enumerate(top_documents):
        documents.extend(_top)
        labels.extend([topic_id] * len(_top))
    indices = list(range(len(documents)))
    shuffle(indices)
    res = ""
    for i_doc in indices:
        res += """
        <div class="document" style="background-color: {bgcolor};">
        {document}
        </div>
        """.format(
            document=prep_document(documents[i_doc]),
            bgcolor=colors[labels[i_doc]],
        )
    return res


def render_widget(
    seeds: list[str],
    topic_names: list[list[str]],
    keywords: list[list[list[str]]],
    topic_descriptions: list[list[str]],
    topic_sizes: list[np.ndarray],
    top_documents: list[list[str]],
) -> str:
    n_topics = 0
    for names in topic_names:
        n_topics += len(names)
    colors = list(COLOR_PALETTE)
    # Colors will loop back when we run out of them
    while len(colors) <= n_topics:
        colors.extend(COLOR_PALETTE)
    res = """<div class="column">\n"""
    res += """
    <div class="button-container">
    <div class="floating-label">Seed Phrases</div>
    """
    for seed_id, seed in enumerate(seeds):
        res += """
        <button id="button-{seed_id}" class="box model-switcher" style="background-color: {bgcolor};"><b>{seed}</b></button>
        """.format(
            bgcolor="rgba(160,188,233, 0.4)" if seed_id == 0 else "white",
            seed_id=seed_id,
            seed=str(seed),
        )
    res += "</div>\n"
    color_start = 0
    for seed_id, seed in enumerate(seeds):
        n_topics = len(topic_names[seed_id])
        seed_colors = colors[color_start : color_start + n_topics]
        color_start += n_topics
        container = """
        <div id="topic-container-{seed_id}" class="box topic-container" style="display: {visibility}">
        <div class="floating-label">Concepts</div>
            {content}
        </div>
        """.format(
            content=render_cards(
                seed_id,
                topic_names[seed_id],
                keywords[seed_id],
                topic_descriptions[seed_id],
                topic_sizes[seed_id],
                seed_colors,
            ),
            visibility="block" if seed_id == 0 else "none",
            seed_id=seed_id,
        )
        res += container
    res += "\n</div>\n"
    color_start = 0
    for seed_id, seed in enumerate(seeds):
        n_topics = len(topic_names[seed_id])
        seed_colors = colors[color_start : color_start + n_topics]
        color_start += n_topics
        res += """
        <div id="document-viewer-{seed_id}" class="box document-viewer" style="display: {visibility};">
        <div class="floating-label">Example Documents</div>
        """.format(
            seed_id=seed_id,
            visibility="flex" if seed_id == 0 else "none",
        )
        res += render_documents(top_documents[seed_id], seed_colors)
        res += "\n</div>\n"
    res += """
    <script>
    const modelSwitchers = Array.from(document.getElementsByClassName('model-switcher'))
    modelSwitchers.forEach(button => {
        console.log("Found this button")
        button.addEventListener('click', function(event) {
            const modelId = button.id.split("-")[1];
            const containers = Array.from(document.getElementsByClassName('topic-container'))
            containers.forEach(container => {
                const containerId = container.id.split("-")[2];
                if (containerId == modelId) {
                    container.style = "display: block";
                } else {
                    container.style = "display: none";
                }
            })
            const documentViewers = Array.from(document.getElementsByClassName('document-viewer'))
            documentViewers.forEach(container => {
                const containerId = container.id.split("-")[2];
                if (containerId == modelId) {
                    container.style = "display: flex";
                } else {
                    container.style = "display: none";
                }
            })
            modelSwitchers.forEach(button => {
                button.style = "background-color: white;";
            })
            button.style = "background-color: rgba(160,188,233, 0.4);";
        });
    });
    </script>
    """
    return res


STYLE = """
body {
    font-family: "Merriweather", Times New Roman;
}
.topic-container {
    max-width: 600px;
    overflow-y: auto;
    overflow-x: hidden;
}
.floating-label {
    padding: 10px;
    position: fixed;
    color: white;
    background-color: black;
    border-radius: 10px;
    z-index: 30;
    width: fit-content;
    margin-top: -30px;
    margin-left: -30px;
}
.card {
    padding-top: 2px;
    padding-bottom: 2px;
    padding-left: 15px;
    padding-right: 15px;
    margin: 5px;
    margin-top: 10px;
    margin-bottom: 10px;
    background-color: #E6F3FF;
    border: solid;
    box-shadow: 0px 0px 1px 1px rgba(0,0,0,0.1);
    border-color: #999999;
    border-width: 1px;
    border-radius: 5px;
    border-color: black;
}
.button-container {
    padding: 10px;
    margin: 30px;
    margin-bottom: 5px;
    background-color: "white";
    border-radius: 5px;
    box-shadow: 0px 0px 1px 1px rgba(0,0,0,0.1);
    align-self: stretch;
    flex-grow: 1;
    flex-shrink: 1;
    display: flex;
    flex-direction: row;
    max-width: 600px;
}
.box {
    padding: 10px;
    margin: 30px;
    background-color: "white";
    border-radius: 5px;
    box-shadow: 0px 0px 1px 1px rgba(0,0,0,0.1);
    align-self: stretch;
    flex-grow: 0;
    flex-shrink: 1;
}
.document {
    padding: 20px;
    margin: 5px;
    border-radius: 5px;
    border: solid;
    box-shadow: 0px 0px 1px 1px rgba(0,0,0,0.1);
    border-color: #999999;
    border-width: 1px;
    max-height: 150px;
    flex-shrink: 0;
    text-align: left;
    text-overflow: ellipsis;
    font-style: italic;
    overflow: hidden;
}
.model-switcher {
    font-size: 16px;
    font-family: "Merriweather", Times New Roman;
    display: flex;
    flex-grow: 1;
    align-items: center;
    align-content: center;
    border: solid;
    border-color: #999999;
    border-width: 1px;
    border-radius: 5px;
    background-color: white;
    margin: 5px;
    color: black;
    padding: 10px;
    justify-content: left;
    text-decoration: none;
}
.model-switcher:hover {
    border-color: black;
}
.document-viewer {
    flex-basis: 600px;
    display: flex;
    justify-content: flex-start;
    flex-direction: column;
    overflow-y: scroll;
    overflow-x: hidden;
}
#container {
    display: flex;
    flex-direction: row;
    flex-grow: 0;
    justify-content: center;
    align-items: stretch;
    align-content: stretch;
    max-height: 1000px;
}
.column {
    display: flex;
    flex-direction: column;
    align-items: stretch;
    flex-basis: fit-content;
}
"""
HTML_WRAPPER = """
<!doctype html>
<html>
  <head>
    <meta charset="UTF-8" />
    <meta name="viewport" content="width=device-width, initial-scale=1.0" />
    <script src="https://cdn.plot.ly/plotly-3.1.0.min.js" charset="utf-8"></script>
    <link rel="stylesheet" href="https://fonts.googleapis.com/css?family=Merriweather">
    <style>
    {style}
    </style>
  </head>
  <body>
    <div id="container">
    {body_content}
    </div>
  </body>
</html>
"""


def create_browser(
    seeds: list[str],
    topic_names: list[list[str]],
    keywords: list[list[list[str]]],
    topic_descriptions: list[list[str]],
    topic_sizes: np.ndarray,
    top_documents: list[list[str]],
) -> HTMLFigure:
    """Creates a concept browser figure with which you can investigate concepts related to different seeds.

    Parameters
    ----------
    seeds: list[str]
        Seed phrases used for the analysis.
    topic_names: list[list[str]]
        Names of the topics for each of the seed phrases.
    keywords: list[list[list[str]]]
        Keywords for each of the topics for each seed.
    topic_descriptions: list[list[str]]
        Descriptions of the topics for each of the seed phrases.
    topic_sizes: np.ndarray
        Sizes of the topics for each seed, preferably number of documents.
    top_documents: list[list[str]]
        Top documents for each of the topics for each seed.

    Returns
    -------
    HTMLFigure
        Interactive HTML figure that you can either display or save.
    """
    html = HTML_WRAPPER.format(
        style=STYLE,
        body_content=render_widget(
            seeds,
            topic_names,
            keywords,
            topic_descriptions,
            topic_sizes,
            top_documents,
        ),
    )
    return HTMLFigure(html)



import itertools
import warnings
from collections import defaultdict
from dataclasses import dataclass
from typing import Optional, Union

import numpy as np
from rich.console import Console
from rich.table import Table
from sentence_transformers import SentenceTransformer
from sklearn.metrics.pairwise import cosine_similarity

from turftopic.base import Encoder
from turftopic.container import TopicContainer
from turftopic.encoders.multimodal import MultimodalEncoder
from turftopic.multimodal import (
    ImageRepr,
    MultimodalEmbeddings,
    _load_images,
    encode_multimodal,
)


def encode_texts(encoder, texts: list[str]) -> np.ndarray:
    if hasattr(encoder, "get_text_embeddings"):
        return encoder.get_text_embeddings(texts)
    return encoder.encode(texts)


def encode_images(encoder, images: list[ImageRepr]):
    images = list(_load_images(images))
    if hasattr(encoder, "get_image_embeddings"):
        return encoder.get_image_embeddings(images)
    return encoder.encode(images)


def _topic_distances(embeddings: np.ndarray, labels: np.ndarray) -> np.ndarray:
    topic_embeddings = [
        embeddings[labels == label] for label in np.unique(labels)
    ]
    topic_dist = []
    for i_topic, j_topic in itertools.combinations(
        np.arange(len(topic_embeddings)), 2
    ):
        dist = 1 - cosine_similarity(
            topic_embeddings[i_topic], topic_embeddings[j_topic]
        )
        topic_dist.append(np.mean(dist))
    return np.array(topic_dist)


def _topic_coherences(
    embeddings: np.ndarray, labels: np.ndarray
) -> np.ndarray:
    sims = []
    for label in np.unique(labels):
        topic_embeddings = embeddings[labels == label]
        sim = cosine_similarity(topic_embeddings, topic_embeddings)
        sim[np.triu_indices(sim.shape[0], 0)] = np.nan
        sims.append(np.nanmean(sim))
    return np.array(sims)


def _topic_coverage(
    desc_embeddings: np.ndarray, document_embeddings: np.ndarray
) -> np.ndarray:
    sim = cosine_similarity(document_embeddings, desc_embeddings)
    # Selecting the maximally similar description item to a document
    max_sim = np.max(sim, axis=1)
    return max_sim


def _encode_descriptives(
    encoder, data: TopicContainer, top_k: int
) -> tuple[dict[str, np.ndarray], dict[str, np.ndarray]]:
    """Encodes descriptive words, documents and images."""
    top_words = data.get_top_words(top_k=top_k)
    top_documents = data.get_top_documents(top_k=top_k)
    embeddings = defaultdict(list)
    labels = defaultdict(list)
    n_topics = len(top_words)
    try:
        classes = data.classes_
    except AttributeError:
        classes = list(range(n_topics))
    for topic_id, words, docs in zip(classes, top_words, top_documents):
        if topic_id == -1:
            # Skipping outlier topic
            continue
        embeddings["words"].extend(encode_texts(encoder, words))
        labels["words"].extend([topic_id] * len(words))
        embeddings["documents"].extend(encode_texts(encoder, docs))
        labels["documents"].extend([topic_id] * len(docs))
    try:
        top_images = data.get_top_images(top_k=top_k)
        for topic_id, images in zip(classes, top_images):
            if topic_id == -1:
                # Skipping outlier topic
                continue
            try:
                embeddings["images"].extend(encode_images(encoder, images))
            except Exception as e:
                warnings.warn(
                    f"Couldn't encode images due to error: {e}, proceeding without image encoding."
                )
            labels["images"].extend([topic_id] * len(images))
    except Exception:
        warnings.warn("No images, proceeding without them.")
        pass
    embeddings = {key: np.array(emb) for key, emb in embeddings.items()}
    labels = {key: np.array(lab) for key, lab in labels.items()}
    return embeddings, labels


@dataclass
class EvaluationResult:
    res: dict[str, dict[str, np.ndarray]]

    @property
    def modalities(self) -> set[str]:
        return set(self.res.keys())

    @property
    def coherence(self) -> float:
        all_scores = []
        for modality in self.res:
            all_scores.extend(self.res[modality]["coherence"])
        return np.mean(all_scores)

    @property
    def diversity(self) -> float:
        all_scores = []
        for modality in self.res:
            all_scores.extend(self.res[modality]["diversity"])
        return np.mean(all_scores)

    @property
    def interpretability(self) -> float:
        return np.sqrt(self.coherence * self.diversity)

    @property
    def coverage(self) -> float:
        all_scores = []
        for modality in self.res:
            all_scores.extend(self.res[modality]["coverage"])
        return np.mean(all_scores)

    def _summary_table(self):
        index = []
        cols = ["coherence", "diversity", "coverage"]
        rows = []
        for modality in self.res:
            row = []
            for metric in cols:
                row.append(np.mean(self.res[modality][metric]))
            rows.append(row)
            index.append(modality)
        return index, cols, rows

    def __str__(self) -> str:
        index, cols, rows = self._summary_table()
        table = Table(show_lines=True)
        table.add_column("Modality")
        for column in cols:
            table.add_column(column.capitalize())
        for ind, row in zip(index, rows):
            row = [f"{val:.2f}" for val in row]
            table.add_row(ind.capitalize(), *row)
        console = Console()
        with console.capture() as capture:
            console.print(table)
        return capture.get()

    def summary_df(self):
        try:
            import pandas as pd
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                "You should install pandas before using dataframe utilities in Turftopic."
            ) from e
        index, cols, rows = self._summary_table()
        return pd.DataFrame(rows, columns=cols, index=index)


class Evaluator:
    def __init__(
        self,
        top_k: int = 10,
        encoder: Union[Encoder, MultimodalEncoder, str] = "all-MiniLM-L6-v2",
    ):
        self.encoder = encoder
        if isinstance(self.encoder, str):
            self.encoder_ = SentenceTransformer(self.encoder)
        else:
            self.encoder_ = self.encoder
        self.top_k = top_k

    def evaluate(
        self,
        data: TopicContainer,
        embeddings: Optional[Union[np.ndarray, MultimodalEmbeddings]] = None,
    ):
        res = defaultdict(dict)
        if embeddings is None:
            if hasattr(data, "images"):
                embeddings = encode_multimodal(
                    self.encoder_, data.corpus, data.images
                )
                document_embeddings = embeddings["document_embeddings"]
            else:
                document_embeddings = encode_texts(self.encoder_, data.corpus)
        elif isinstance(embeddings, dict):
            document_embeddings = embeddings["document_embeddings"]
        else:
            document_embeddings = embeddings
        desc_embeddings, labels = _encode_descriptives(
            self.encoder_, data, self.top_k
        )
        for modality in desc_embeddings:
            coherence = _topic_coherences(
                desc_embeddings[modality], labels[modality]
            )
            res[modality]["coherence"] = coherence
            diversity = _topic_distances(
                desc_embeddings[modality], labels[modality]
            )
            res[modality]["diversity"] = diversity
            coverage = _topic_coverage(
                desc_embeddings[modality], document_embeddings
            )
            res[modality]["coverage"] = coverage
        return EvaluationResult(res)


def compare_models(results: list[EvaluationResult]):
    import pandas as pd
    import statsmodels.formula.api as smf

    modalities = set(results[0])

    records = []
    for modality in ["words", "documents"]:
        for i, _res in enumerate(eval_res):
            for score in _res.res[modality]["coverage"]:
                records.append(
                    dict(model=f"Model {i}", modality=modality, score=score)
                )
    df = pd.DataFrame.from_records(records)
    rand_eff_mod = smf.ols(
        formula="score ~ C(model)",
        data=df,
        groups=df["modality"],
    )
    reg_res = rand_eff_mod.fit()
    reg_res.summary()



from typing import Optional

from transformers import pipeline

from turftopic.analyzers.base import Analyzer

T5_NAME_PROMPT = "What is the topic? Keywords: {keywords}"
T5_DESC_PROMPT = "Summarize in one sentence. Keywords: {keywords}"


class T5Analyzer(Analyzer):
    """Analyze topic model with a text-to-text model.

    Parameters
    ----------
    model_name: str, default 'google/flan-t5-small'
        Text-to-text model to use for analyses.
    use_summaries: bool, default False
        Indicates whether the language model should summarize documents before
        analyzing the topics.
    context: str, default None
        Additional context provided to the analyzer for analysis.
        e.g. "Analyze topics from blog posts related to morality and religion"
    system_prompt: str, default None
        Ignored, exists for compatibility
    summary_prompt: str, default None
        Prompt to use for abstractive summarization.
    namer_prompt: str, default None
        Prompt template for naming topics.
    description_prompt: str, default None
        Prompt template for generating topic descriptions.
    device: str, default "cpu"
        ID of the device to run the language model on.
    """

    def __init__(
        self,
        model_name: str = "google/flan-t5-small",
        context: Optional[str] = None,
        use_summaries: bool = False,
        system_prompt: Optional[str] = None,
        summary_prompt: Optional[str] = None,
        namer_prompt: Optional[str] = T5_NAME_PROMPT,
        description_prompt: Optional[str] = T5_DESC_PROMPT,
        device: str = "cpu",
    ):
        self.device = device
        self.model_name = model_name
        self.pipeline = pipeline(
            task="text2text-generation",
            model=self.model_name,
            device=self.device,
        )
        self.summary_prompt = summary_prompt or self.summary_prompt
        self.namer_prompt = namer_prompt or self.namer_prompt
        self.description_prompt = description_prompt or self.description_prompt
        self.use_summaries = use_summaries

    def generate_text(self, prompt: str) -> str:
        return self.pipeline(prompt)



import random
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn


def pairwise_euclidean_distance(x, y):
    cost = (
        torch.sum(x**2, axis=1, keepdim=True)
        + torch.sum(y**2, dim=1)
        - 2 * torch.matmul(x, y.t())
    )
    return cost


class ETP(nn.Module):
    def __init__(
        self,
        sinkhorn_alpha,
        init_a_dist=None,
        init_b_dist=None,
        OT_max_iter=5000,
        stopThr=0.5e-2,
    ):
        super().__init__()
        self.sinkhorn_alpha = sinkhorn_alpha
        self.OT_max_iter = OT_max_iter
        self.stopThr = stopThr
        self.epsilon = 1e-16
        self.init_a_dist = init_a_dist
        self.init_b_dist = init_b_dist
        if init_a_dist is not None:
            self.a_dist = init_a_dist
        if init_b_dist is not None:
            self.b_dist = init_b_dist

    def forward(self, x, y):
        # Sinkhorn's algorithm
        M = pairwise_euclidean_distance(x, y)
        device = M.device
        if self.init_a_dist is None:
            a = (torch.ones(M.shape[0]) / M.shape[0]).unsqueeze(1).to(device)
        else:
            a = F.softmax(self.a_dist, dim=0).to(device)
        if self.init_b_dist is None:
            b = (torch.ones(M.shape[1]) / M.shape[1]).unsqueeze(1).to(device)
        else:
            b = F.softmax(self.b_dist, dim=0).to(device)
        u = (torch.ones_like(a) / a.size()[0]).to(device)  # Kx1
        K = torch.exp(-M * self.sinkhorn_alpha)
        err = 1
        cpt = 0
        while err > self.stopThr and cpt < self.OT_max_iter:
            v = torch.div(b, torch.matmul(K.t(), u) + self.epsilon)
            u = torch.div(a, torch.matmul(K, v) + self.epsilon)
            cpt += 1
            if cpt % 50 == 1:
                bb = torch.mul(v, torch.matmul(K.t(), u))
                err = torch.norm(
                    torch.sum(torch.abs(bb - b), dim=0), p=float("inf")
                )
        transp = u * (K * v.T)
        loss_ETP = torch.sum(transp * M)
        return loss_ETP, transp


class fastopic(nn.Module):
    def __init__(
        self,
        num_topics: int,
        theta_temp: float = 1.0,
        DT_alpha: float = 3.0,
        TW_alpha: float = 2.0,
        random_state: Optional[int] = None,
    ):
        super().__init__()

        self.num_topics = num_topics
        self.DT_alpha = DT_alpha
        self.TW_alpha = TW_alpha
        self.theta_temp = theta_temp
        self.seed = random_state or random.randint(0, 10_000)
        self.epsilon = 1e-12

    def init(self, vocab_size: int, embed_size: int):
        torch.manual_seed(self.seed)
        self.word_embeddings = nn.init.trunc_normal_(
            torch.empty(vocab_size, embed_size)
        )
        self.word_embeddings = nn.Parameter(F.normalize(self.word_embeddings))
        self.topic_embeddings = torch.empty((self.num_topics, embed_size))
        nn.init.trunc_normal_(self.topic_embeddings, std=0.1)
        self.topic_embeddings = nn.Parameter(
            F.normalize(self.topic_embeddings)
        )
        self.word_weights = nn.Parameter(
            (torch.ones(vocab_size) / vocab_size).unsqueeze(1)
        )
        self.topic_weights = nn.Parameter(
            (torch.ones(self.num_topics) / self.num_topics).unsqueeze(1)
        )
        self.DT_ETP = ETP(self.DT_alpha, init_b_dist=self.topic_weights)
        self.TW_ETP = ETP(self.TW_alpha, init_b_dist=self.word_weights)

    def get_transp_DT(
        self,
        doc_embeddings,
    ):
        torch.manual_seed(self.seed)
        topic_embeddings = self.topic_embeddings.detach().to(
            doc_embeddings.device
        )
        _, transp = self.DT_ETP(doc_embeddings, topic_embeddings)
        return transp.detach().cpu().numpy()

    # only for testing
    def get_beta(self):
        torch.manual_seed(self.seed)
        _, transp_TW = self.TW_ETP(self.topic_embeddings, self.word_embeddings)
        # use transport plan as beta
        beta = transp_TW * transp_TW.shape[0]
        return beta

    # only for testing
    def get_theta(self, doc_embeddings, train_doc_embeddings):
        torch.manual_seed(self.seed)
        topic_embeddings = self.topic_embeddings.detach().to(
            doc_embeddings.device
        )
        dist = pairwise_euclidean_distance(doc_embeddings, topic_embeddings)
        train_dist = pairwise_euclidean_distance(
            train_doc_embeddings, topic_embeddings
        )
        exp_dist = torch.exp(-dist / self.theta_temp)
        exp_train_dist = torch.exp(-train_dist / self.theta_temp)
        theta = exp_dist / (exp_train_dist.sum(0))
        theta = theta / theta.sum(1, keepdim=True)
        return theta

    def forward(self, train_bow, doc_embeddings):
        torch.manual_seed(self.seed)
        loss_DT, transp_DT = self.DT_ETP(doc_embeddings, self.topic_embeddings)
        loss_TW, transp_TW = self.TW_ETP(
            self.topic_embeddings, self.word_embeddings
        )
        loss_ETP = loss_DT + loss_TW
        theta = transp_DT * transp_DT.shape[0]
        beta = transp_TW * transp_TW.shape[0]
        # Dual Semantic-relation Reconstruction
        recon = torch.matmul(theta, beta)
        loss_DSR = (
            -(train_bow * (recon + self.epsilon).log()).sum(axis=1).mean()
        )
        loss = loss_DSR + loss_ETP
        rst_dict = {
            "loss": loss,
        }
        return rst_dict



import warnings

from turftopic.analyzers import LLMAnalyzer, OpenAIAnalyzer

warnings.warn(
    """
The turftopic.namers is deprecated in favor of turftopic.analyzers and will be fully phased out in 1.1.0!
For now, we provide aliases to the analyzers module, but we encourage you to use the analyzers module
directly. """,
    DeprecationWarning,
)

LLMTopicNamer = LLMAnalyzer
OpenAITopicNamer = OpenAIAnalyzer

__all__ = ["OpenAITopicNamer", "LLMTopicNamer"]



import warnings
from datetime import datetime
from typing import Literal, Optional, Union

import numpy as np
import scipy.sparse as spr
from PIL import Image
from rich.console import Console
from sentence_transformers import SentenceTransformer
from sklearn.exceptions import NotFittedError
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.preprocessing import normalize

from turftopic.base import ContextualModel, Encoder
from turftopic.data import TopicData
from turftopic.dynamic import DynamicTopicModel
from turftopic.encoders.multimodal import MultimodalEncoder
from turftopic.hierarchical import DivisibleTopicNode
from turftopic.models._keynmf import KeywordNMF, SBertKeywordExtractor
from turftopic.models.wnmf import weighted_nmf
from turftopic.multimodal import (
    ImageRepr,
    MultimodalEmbeddings,
    MultimodalModel,
)
from turftopic.vectorizers.default import default_vectorizer


class KeyNMF(ContextualModel, DynamicTopicModel, MultimodalModel):
    """Extracts keywords from documents based on semantic similarity of
    term encodings to document encodings.
    Topics are then extracted with non-negative matrix factorization from
    keywords' proximity to documents.

    ```python
    from turftopic import KeyNMF

    corpus: list[str] = ["some text", "more text", ...]

    model = KeyNMF(10, top_n=10).fit(corpus)
    model.print_topics()
    ```

    Parameters
    ----------
    n_components: int or "auto"
        Number of topics.
        Auto selects the number of topics using
        the Bayesian Information Criterion.
    encoder: str or SentenceTransformer
        Model to encode documents/terms, all-MiniLM-L6-v2 is the default.
    vectorizer: CountVectorizer, default None
        Vectorizer used for term extraction.
        Can be used to prune or filter the vocabulary.
    top_n: int, default 25
        Number of keywords to extract for each document.
    random_state: int, default None
        Random state to use so that results are exactly reproducible.
    metric: "cosine" or "dot", default "cosine"
        Similarity metric to use for keyword extraction.
    seed_phrase: str, default None
        Describes an aspect of the corpus that the model should explore.
        It can be a free-text query, such as
        "Christian Denominations: Protestantism and Catholicism"
    seed_exponent: float, default 2.0
        Exponent that is applied to document weight in relation to the provided seed phrase.
    cross_lingual: bool, default False
        Indicates whether KeyNMF should match terms across languages.
        This is useful when you have a corpus containing multiple languages.
    term_match_threshold: float, default 0.9
        Cosine similarity threshold for matching terms across languages.
    """

    def __init__(
        self,
        n_components: Union[int, Literal["auto"]],
        encoder: Union[
            Encoder, str, MultimodalEncoder
        ] = "sentence-transformers/all-MiniLM-L6-v2",
        vectorizer: Optional[CountVectorizer] = None,
        top_n: int = 25,
        random_state: Optional[int] = None,
        metric: Literal["cosine", "dot"] = "cosine",
        seed_phrase: Optional[str] = None,
        seed_exponent: float = 2.0,
        cross_lingual: bool = False,
        term_match_threshold: float = 0.9,
    ):
        self.random_state = random_state
        self.n_components = n_components
        self.top_n = top_n
        self.seed_exponent = seed_exponent
        self.metric = metric
        self.encoder = encoder
        self._has_custom_vectorizer = vectorizer is not None
        if isinstance(encoder, str):
            if (
                encoder == "sentence-transformers/all-MiniLM-L6-v2"
            ) and cross_lingual:
                warnings.warn(
                    "all-MiniLM is incompatible with cross-lingual transfer, using paraphrase-multilingual-MiniLM-L12-v2."
                )
                encoder = "paraphrase-multilingual-MiniLM-L12-v2"
            self.encoder_ = SentenceTransformer(encoder)
        else:
            self.encoder_ = encoder
        self.validate_encoder()
        if vectorizer is None:
            self.vectorizer = default_vectorizer()
        else:
            self.vectorizer = vectorizer
        self.model = KeywordNMF(
            n_components=n_components, seed=random_state, top_n=self.top_n
        )
        self.extractor = SBertKeywordExtractor(
            top_n=self.top_n,
            vectorizer=self.vectorizer,
            encoder=self.encoder_,
            metric=self.metric,
        )
        self.seed_phrase = seed_phrase
        self.seed_embedding = None
        if self.seed_phrase is not None:
            self.seed_embedding = self.encoder_.encode([self.seed_phrase])[0]
        self.cross_lingual = cross_lingual
        self.term_match_threshold = term_match_threshold

    def extract_keywords(
        self,
        batch_or_document: Union[str, list[str]],
        embeddings: Optional[np.ndarray] = None,
        fitting: bool = True,
    ) -> list[dict[str, float]]:
        """Extracts keywords from a document or a batch of documents.

        Parameters
        ----------
        batch_or_document: str | list[str]
            A single document or a batch of documents.
        embeddings: ndarray, optional
            Precomputed document embeddings.
        """
        if isinstance(batch_or_document, str):
            batch_or_document = [batch_or_document]
        keywords = self.extractor.batch_extract_keywords(
            batch_or_document,
            embeddings=embeddings,
            seed_embedding=self.seed_embedding,
            seed_exponent=self.seed_exponent,
            fitting=fitting,
        )
        if self.cross_lingual:
            keywords = self.extractor.match_terms(
                keywords, threshold=self.term_match_threshold
            )
        return keywords

    def vectorize(
        self,
        raw_documents=None,
        embeddings: Optional[np.ndarray] = None,
        keywords: Optional[list[dict[str, float]]] = None,
    ) -> spr.csr_array:
        """Creates document-term-matrix from documents."""
        if keywords is None:
            keywords = self.extract_keywords(
                raw_documents, embeddings=embeddings
            )
        return self.model.vectorize(keywords)

    def divide_topic(
        self,
        node: DivisibleTopicNode,
        n_subtopics: int,
    ) -> list[DivisibleTopicNode]:
        document_term_matrix = getattr(self, "document_term_matrix", None)
        if document_term_matrix is None:
            raise ValueError(
                "document_term_matrix is needed for computing hierarchies. Perhaps you fitted the model online?"
            )
        dtm = document_term_matrix
        subtopics = []
        weight = node.document_topic_vector
        subcomponents, sub_doc_topic = weighted_nmf(
            dtm, weight, n_subtopics, self.random_state, max_iter=200
        )
        subcomponents = subcomponents * np.log(
            1 + subcomponents.mean() / (subcomponents.sum(axis=0) + 1)
        )
        subcomponents = normalize(subcomponents, axis=1, norm="l2")
        for i, component, doc_topic_vector in zip(
            range(n_subtopics), subcomponents, sub_doc_topic.T
        ):
            sub = DivisibleTopicNode(
                self,
                path=(*node.path, i),
                word_importance=component,
                document_topic_vector=doc_topic_vector,
                children=None,
            )
            subtopics.append(sub)
        return subtopics

    def fit_transform(
        self,
        raw_documents=None,
        y=None,
        embeddings: Optional[np.ndarray] = None,
        keywords: Optional[list[dict[str, float]]] = None,
    ) -> np.ndarray:
        """Fits topic model and returns topic importances for documents.

        Parameters
        ----------
        raw_documents: iterable of str, optional
            Documents to fit the model on.
        embeddings: ndarray of shape (n_documents, n_dimensions), optional
            Precomputed document encodings.
        keywords: list[dict[str, float]], optional
            Precomputed keyword dictionaries.

        Returns
        -------
        ndarray of shape (n_dimensions, n_topics)
            Document-topic matrix.
        """
        if self.seed_phrase is not None and keywords is not None:
            warnings.warn(
                "Seed phrase is specified, but keyword matrix is pre-computed. The seed phrase will be ignored. Note that this is not a problem if you already calculated the keyword matrix using the seed phrase."
            )
        console = Console()
        with console.status("Running KeyNMF") as status:
            if keywords is None:
                status.update("Extracting keywords")
                keywords = self.extract_keywords(
                    raw_documents, embeddings=embeddings
                )
                console.log("Keyword extraction done.")
            status.update("Decomposing with NMF")
            try:
                doc_topic_matrix = self.model.transform(keywords)
            except (NotFittedError, AttributeError):
                doc_topic_matrix = self.model.fit_transform(keywords)
                self.components_ = self.model.components
            console.log("Model fitting done.")
        self.document_topic_matrix = doc_topic_matrix
        self.document_term_matrix = self.model.vectorize(keywords)
        self.hierarchy = DivisibleTopicNode.create_root(
            self, self.components_, self.document_topic_matrix
        )
        self.top_documents = self.get_top_documents(
            raw_documents, document_topic_matrix=doc_topic_matrix
        )
        return doc_topic_matrix

    def fit_transform_multimodal(
        self,
        raw_documents: list[str],
        images: list[ImageRepr],
        y=None,
        embeddings: Optional[MultimodalEmbeddings] = None,
    ) -> np.ndarray:
        self.validate_embeddings(embeddings)
        console = Console()
        self.multimodal_embeddings = embeddings
        with console.status("Fitting model") as status:
            if self.multimodal_embeddings is None:
                status.update("Encoding documents")
                self.multimodal_embeddings = self.encode_multimodal(
                    raw_documents, images
                )
                console.log("Documents encoded.")
            status.update("Extracting keywords")
            document_keywords = self.extract_keywords(
                raw_documents,
                embeddings=self.multimodal_embeddings["document_embeddings"],
            )
            image_keywords = self.extract_keywords(
                raw_documents,
                embeddings=self.multimodal_embeddings["image_embeddings"],
            )
            console.log("Keyword extraction done.")
            status.update("Decomposing with NMF")
            try:
                doc_topic_matrix = self.model.transform(document_keywords)
            except (NotFittedError, AttributeError):
                doc_topic_matrix = self.model.fit_transform(document_keywords)
                self.components_ = self.model.components
            console.log("Model fitting done.")
            status.update("Transforming images")
            self.image_topic_matrix = self.model.transform(image_keywords)
            self.top_images: list[list[Image.Image]] = self.collect_top_images(
                images, self.image_topic_matrix
            )
            console.log("Images transformed")
        self.document_topic_matrix = doc_topic_matrix
        self.document_term_matrix = self.model.vectorize(document_keywords)
        self.hierarchy = DivisibleTopicNode.create_root(
            self, self.components_, self.document_topic_matrix
        )
        self.top_documents = self.get_top_documents(
            raw_documents, document_topic_matrix=doc_topic_matrix
        )
        return doc_topic_matrix

    def fit(
        self,
        raw_documents=None,
        y=None,
        embeddings: Optional[np.ndarray] = None,
        keywords: Optional[list[dict[str, float]]] = None,
    ) -> np.ndarray:
        """Fits topic model and returns topic importances for documents.

        Parameters
        ----------
        raw_documents: iterable of str, optional
            Documents to fit the model on.
        embeddings: ndarray of shape (n_documents, n_dimensions), optional
            Precomputed document encodings.
        keywords: list[dict[str, float]], optional
            Precomputed keyword dictionaries.
        """
        self.fit_transform(
            raw_documents, y, embeddings=embeddings, keywords=keywords
        )
        return self

    def get_vocab(self) -> np.ndarray:
        return np.array(self.model.index_to_key)

    def transform(
        self,
        raw_documents=None,
        embeddings: Optional[np.ndarray] = None,
        keywords: Optional[list[dict[str, float]]] = None,
    ) -> np.ndarray:
        """Infers topic importances for new documents based on a fitted model.

        Parameters
        ----------
        raw_documents: iterable of str
            Documents to fit the model on.
        embeddings: ndarray of shape (n_documents, n_dimensions), optional
            Precomputed document encodings.
        keywords: list[dict[str, float]], optional
            Precomputed keyword dictionaries.

        Returns
        -------
        ndarray of shape (n_dimensions, n_topics)
            Document-topic matrix.
        """
        if self.seed_phrase is not None and keywords is not None:
            warnings.warn(
                "Seed phrase is specified, but keyword matrix is pre-computed. The seed phrase will be ignored. Note that this is not a problem if you already calculated the keyword matrix using the seed phrase."
            )
        if keywords is None and raw_documents is None:
            raise ValueError(
                "You have to pass either keywords or raw_documents."
            )
        if keywords is None:
            keywords = self.extract_keywords(
                list(raw_documents),
                embeddings=embeddings,
                fitting=False,
            )
        return self.model.transform(keywords)

    def partial_fit(
        self,
        raw_documents: Optional[list[str]] = None,
        embeddings: Optional[np.ndarray] = None,
        keywords: Optional[list[dict[str, float]]] = None,
    ):
        """Online fits KeyNMF on a batch of documents.

        Parameters
        ----------
        raw_documents: iterable of str
            Documents to fit the model on.
        embeddings: ndarray of shape (n_documents, n_dimensions), optional
            Precomputed document encodings.
        keywords: list[dict[str, float]], optional
            Precomputed keyword dictionaries.
        """
        if self.seed_phrase is not None and keywords is not None:
            warnings.warn(
                "Seed phrase is specified, but keyword matrix is pre-computed. The seed phrase will be ignored. Note that this is not a problem if you already calculated the keyword matrix using the seed phrase."
            )
        if self.cross_lingual:
            raise ValueError(
                "Cross-lingual online topic modeling is yet to be implemented in KeyNMF."
            )
        if not self._has_custom_vectorizer:
            self.vectorizer = CountVectorizer(stop_words="english")
            self._has_custom_vectorizer = True
        min_df = self.vectorizer.min_df
        max_df = self.vectorizer.max_df
        if (min_df != 1) or (max_df != 1.0):
            warnings.warn(
                f"""When applying partial fitting, the vectorizer is fitted batch-wise in KeyNMF.
            You have a vectorizer with min_df={min_df}, and max_df={max_df}.
            If you continue with these settings, all tokens might get filtered out.
            We recommend setting min_df=1 and max_df=1.0 for online fitting.
            `model = KeyNMF(10, vectorizer=CountVectorizer(min_df=1, max_df=1.0)`
            """
            )
        if keywords is None and raw_documents is None:
            raise ValueError(
                "You have to pass either keywords or raw_documents."
            )
        if keywords is None:
            keywords = self.extract_keywords(
                raw_documents, embeddings=embeddings
            )
        self.model.partial_fit(keywords)
        self.components_ = self.model.components
        return self

    def prepare_topic_data(
        self,
        corpus: Optional[list[str]],
        embeddings: Optional[np.ndarray] = None,
        keywords: Optional[list[dict[str, float]]] = None,
    ) -> TopicData:
        if keywords is None and corpus is None:
            raise ValueError(
                "You have to pass either keywords or raw_documents."
            )
        if self.seed_phrase is not None and keywords is not None:
            warnings.warn(
                "Seed phrase is specified, but keyword matrix is pre-computed. The seed phrase will be ignored. Note that this is not a problem if you already calculated the keyword matrix using the seed phrase."
            )
        console = Console()
        with console.status("Running KeyNMF") as status:
            if embeddings is None:
                embeddings = self.encode_documents(corpus)
            if keywords is None:
                status.update("Extracting keywords")
                keywords = self.extract_keywords(corpus, embeddings=embeddings)
                console.log("Keyword extraction done.")
            if (corpus is not None) and (len(keywords) != len(corpus)):
                raise ValueError(
                    "length of keywords is not the same as length of the corpus"
                )
            status.update("Decomposing with NMF")
            try:
                doc_topic_matrix = self.model.transform(keywords)
            except (NotFittedError, AttributeError):
                doc_topic_matrix = self.model.fit_transform(keywords)
                self.components_ = self.model.components
            console.log("Model fitting done.")
            document_term_matrix = self.model.vectorize(keywords)
        self.document_topic_matrix = doc_topic_matrix
        self.document_term_matrix = document_term_matrix
        self.hierarchy = DivisibleTopicNode.create_root(
            self, self.components_, self.document_topic_matrix
        )
        res = TopicData(
            corpus=corpus,
            document_term_matrix=document_term_matrix,
            vocab=self.get_vocab(),
            document_topic_matrix=doc_topic_matrix,
            document_representation=embeddings,
            topic_term_matrix=self.components_,  # type: ignore
            transform=getattr(self, "transform", None),
            topic_names=self.topic_names,
            hierarchy=getattr(self, "hierarchy", None),
        )
        return res

    def prepare_dynamic_topic_data(
        self,
        corpus: Optional[list[str]],
        timestamps: list[datetime],
        embeddings: Optional[np.ndarray] = None,
        bins: Union[int, list[datetime]] = 10,
        keywords: Optional[list[dict[str, float]]] = None,
    ):
        if self.seed_phrase is not None and keywords is not None:
            warnings.warn(
                "Seed phrase is specified, but keyword matrix is pre-computed. The seed phrase will be ignored. Note that this is not a problem if you already calculated the keyword matrix using the seed phrase."
            )
        if ((corpus is not None) and (keywords is not None)) and (
            len(keywords) != len(corpus)
        ):
            raise ValueError(
                "Length of keywords is not the same as length of the corpus"
            )
        if (corpus is None) and (keywords is None):
            raise TypeError(
                "You have to pass keywords or a corpus, but both are None."
            )
        if embeddings is None:
            embeddings = self.encode_documents(corpus)
        console = Console()
        with console.status("Running KeyNMF") as status:
            if embeddings is None:
                embeddings = self.encode_documents(corpus)
            if keywords is None:
                status.update("Extracting keywords")
                keywords = self.extract_keywords(corpus, embeddings=embeddings)
                console.log("Keyword extraction done.")
            if (corpus is not None) and (len(keywords) != len(corpus)):
                raise ValueError(
                    "length of keywords is not the same as length of the corpus"
                )
            status.update("Decomposing with NMF")
            try:
                doc_topic_matrix = self.model.transform(keywords)
            except (NotFittedError, AttributeError):
                doc_topic_matrix = self.fit_transform_dynamic(
                    corpus,
                    embeddings=embeddings,
                    keywords=keywords,
                    bins=bins,
                    timestamps=timestamps,
                )
            console.log("Model fitting done.")
            document_term_matrix = self.model.vectorize(keywords)
        try:
            classes = self.classes_
        except AttributeError:
            classes = list(range(self.components_.shape[0]))
        res = TopicData(
            corpus=corpus,
            document_term_matrix=document_term_matrix,
            vocab=self.get_vocab(),
            document_topic_matrix=doc_topic_matrix,
            document_representation=embeddings,
            topic_term_matrix=self.components_,  # type: ignore
            transform=getattr(self, "transform", None),
            topic_names=self.topic_names,
            classes=classes,
            temporal_components=self.temporal_components_,
            temporal_importance=self.temporal_importance_,
            time_bin_edges=self.time_bin_edges,
        )
        return res

    def fit_transform_dynamic(
        self,
        raw_documents=None,
        timestamps: Optional[list[datetime]] = None,
        embeddings: Optional[np.ndarray] = None,
        keywords: Optional[list[dict[str, float]]] = None,
        bins: Union[int, list[datetime]] = 10,
    ) -> np.ndarray:
        if self.seed_phrase is not None and keywords is not None:
            warnings.warn(
                "Seed phrase is specified, but keyword matrix is pre-computed. The seed phrase will be ignored. Note that this is not a problem if you already calculated the keyword matrix using the seed phrase."
            )
        if timestamps is None:
            raise TypeError(
                "You have to pass timestamps when fitting a dynamic model."
            )
        if keywords is None and raw_documents is None:
            raise ValueError(
                "You have to pass either keywords or raw_documents."
            )
        if keywords is None:
            keywords = self.extract_keywords(
                raw_documents, embeddings=embeddings
            )
        time_labels, self.time_bin_edges = self.bin_timestamps(
            timestamps, bins
        )
        doc_topic_matrix = self.model.fit_transform_dynamic(
            keywords, time_labels, self.time_bin_edges
        )
        self.temporal_importance_ = (
            self.model.temporal_importance_.T
            / self.model.temporal_importance_.sum(axis=1)
        ).T
        self.temporal_components_ = self.model.temporal_components
        self.components_ = self.model.components
        self.document_topic_matrix = doc_topic_matrix
        self.document_term_matrix = self.model.vectorize(keywords)
        self.hierarchy = DivisibleTopicNode.create_root(
            self, self.components_, self.document_topic_matrix
        )
        self.top_documents = self.get_top_documents(
            raw_documents, document_topic_matrix=doc_topic_matrix
        )
        return doc_topic_matrix

    def partial_fit_dynamic(
        self,
        raw_documents=None,
        timestamps: Optional[list[datetime]] = None,
        embeddings: Optional[np.ndarray] = None,
        keywords: Optional[list[dict[str, float]]] = None,
        bins: Union[int, list[datetime]] = 10,
    ):
        """Online fits Dynamic KeyNMF on a batch of documents.

        Parameters
        ----------
        raw_documents: iterable of str
            Documents to fit the model on.
        embeddings: ndarray of shape (n_documents, n_dimensions), optional
            Precomputed document encodings.
        keywords: list[dict[str, float]], optional
            Precomputed keyword dictionaries.
        timestamps: list[datetime], optional
            List of timestamps for the batch.
        bins: list[datetime]
            Explicit time bin edges for the dynamic model.
        """
        if self.seed_phrase is not None and keywords is not None:
            warnings.warn(
                "Seed phrase is specified, but keyword matrix is pre-computed. The seed phrase will be ignored. Note that this is not a problem if you already calculated the keyword matrix using the seed phrase."
            )
        if self.cross_lingual:
            raise ValueError(
                "Cross-lingual online topic modeling is yet to be implemented in KeyNMF."
            )
        if timestamps is None:
            raise TypeError(
                "You have to pass timestamps when fitting a dynamic model."
            )
        if keywords is None and raw_documents is None:
            raise ValueError(
                "You have to pass either keywords or raw_documents."
            )
        time_bin_edges = getattr(self, "time_bin_edges", None)
        if time_bin_edges is None:
            if isinstance(bins, int):
                raise TypeError(
                    "You have to pass explicit time bins (list of time bin edges) when partial "
                    "fitting KeyNMF, at least at the first call."
                )
            else:
                self.time_bin_edges = bins
        time_labels, self.time_bin_edges = self.bin_timestamps(
            timestamps, self.time_bin_edges
        )
        if keywords is None:
            keywords = self.extract_keywords(
                raw_documents, embeddings=embeddings
            )
        self.model.partial_fit_dynamic(
            keywords, time_labels, self.time_bin_edges
        )
        self.temporal_importance_ = (
            self.model.temporal_importance_.T
            / self.model.temporal_importance_.sum(axis=1)
        ).T
        self.temporal_components_ = self.model.temporal_components
        self.components_ = self.model.components
        return self



import itertools
from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from typing import Any, Iterable, Optional, Union

import numpy as np
from sklearn.exceptions import NotFittedError

from turftopic.data import TopicData


def bin_timestamps(
    timestamps: list[datetime],
    bins: Union[int, list[datetime]] = 10,
) -> tuple[np.ndarray, list[datetime]]:
    if not len(timestamps) and not isinstance(timestamps[0], datetime):
        raise TypeError("Timestamps have to be `datetime` objects.")
    unix_timestamps = [timestamp.timestamp() for timestamp in timestamps]
    if isinstance(bins, list):
        if min(timestamps) < min(bins):
            raise ValueError(
                f"Earliest timestamp ({min(timestamps)}) is not later or the same as first bin edge ({min(bins)})."
            )
        if max(timestamps) >= max(bins):
            raise ValueError(
                f"Latest timestamp ({max(timestamps)}) is not earlier than last bin edge ({max(bins)})."
            )
        unix_bins = [bin.timestamp() for bin in bins]
        # Have to substract one, else it starts from one
        return np.digitize(unix_timestamps, unix_bins) - 1, bins
    else:
        # Adding one day, so that the maximum value is still included.
        max_timestamp = max(timestamps) + timedelta(days=1)
        unix_bins = np.histogram_bin_edges(unix_timestamps, bins=bins)
        unix_bins[-1] = max_timestamp.timestamp()
        bins = [datetime.fromtimestamp(ts) for ts in unix_bins]
        # Have to substract one, else it starts from one
        return np.digitize(unix_timestamps, unix_bins) - 1, bins


class DynamicTopicModel(ABC):
    @staticmethod
    def bin_timestamps(
        timestamps: list[datetime], bins: Union[int, list[datetime]] = 10
    ) -> tuple[np.ndarray, list[datetime]]:
        """Bins timestamps based on given bins.

        Parameters
        ----------
        timestamps: list[datetime]
            List of timestamps for documents.
        bins: int or list[datetime], default 10
            Time bins to use.
            If the bins are an int (N), N equally sized bins are used.
            Otherwise they should be bin edges, including the last and first edge.
            Bins are inclusive at the lower end and exclusive at the upper (lower <= timestamp < upper).

        Returns
        -------
        time_labels: ndarray of int
            Labels for time slice in each document.
        bin_edges: list[datetime]
            List of edges for time bins.
        """
        return bin_timestamps(timestamps, bins)

    @abstractmethod
    def fit_transform_dynamic(
        self,
        raw_documents,
        timestamps: list[datetime],
        embeddings: Optional[np.ndarray] = None,
        bins: Union[int, list[datetime]] = 10,
    ) -> np.ndarray:
        """Fits a dynamic topic model on the corpus and returns document-topic-importances.

        Parameters
        ----------
        raw_documents
            Documents to fit the model on.
        timestamps: list[datetime]
            Timestamp for each document in `datetime` format.
        embeddings: np.ndarray, default None
            Document embeddings produced by an embedding model.
        bins: int or list[datetime], default 10
            Specifies how to bin timestamps in to time slices.
            When an `int`, the corpus will be divided into N equal time slices.
            When a list, it describes the edges of each time slice including the starting
            and final edges of the slices.

        Returns
        -------
        ndarray of shape (n_documents, n_topics)
            Document-topic importance matrix.
        """
        pass

    def fit_dynamic(
        self,
        raw_documents,
        timestamps: list[datetime],
        embeddings: Optional[np.ndarray] = None,
        bins: Union[int, list[datetime]] = 10,
    ):
        """Fits a dynamic topic model on the corpus and returns document-topic-importances.

        Parameters
        ----------
        raw_documents
            Documents to fit the model on.
        timestamps: list[datetime]
            Timestamp for each document in `datetime` format.
        embeddings: np.ndarray, default None
            Document embeddings produced by an embedding model.
        bins: int or list[datetime], default 10
            Specifies how to bin timestamps in to time slices.
            When an `int`, the corpus will be divided into N equal time slices.
            When a list, it describes the edges of each time slice including the starting
            and final edges of the slices.

            Note: The final edge is not included. You might want to add one day to
            the last bin edge if it equals the last timestamp.
        """
        self.fit_transform_dynamic(raw_documents, timestamps, embeddings, bins)
        return self

    def prepare_dynamic_topic_data(
        self,
        corpus: list[str],
        timestamps: list[datetime],
        embeddings: Optional[np.ndarray] = None,
        bins: Union[int, list[datetime]] = 10,
    ):
        """Produces topic inference data for a given corpus, that can be then used and reused.
        Exists to allow visualizations out of the box with topicwizard.

        Parameters
        ----------
        corpus: list of str
            Documents to infer topical content for.
        timestamps: list[datetime]
            Timestamp for each document in `datetime` format.
        embeddings: ndarray of shape (n_documents, n_dimensions)
            Embeddings of documents.
        bins: int or list[datetime], default 10
            Specifies how to bin timestamps in to time slices.
            When an `int`, the corpus will be divided into N equal time slices.
            When a list, it describes the edges of each time slice including the starting
            and final edges of the slices.

            Note: The final edge is not included. You might want to add one day to
            the last bin edge if it equals the last timestamp.

        Returns
        -------
        TopicData
            Information about topical inference in a dictionary.
        """
        if embeddings is None:
            embeddings = self.encode_documents(corpus)
        if getattr(self, "temporal_components_", None) is not None:
            try:
                document_topic_matrix = self.transform(
                    corpus, embeddings=embeddings
                )
            except (AttributeError, NotFittedError):
                document_topic_matrix = self.fit_transform_dynamic(
                    corpus,
                    timestamps=timestamps,
                    embeddings=embeddings,
                    bins=bins,
                )
        else:
            document_topic_matrix = self.fit_transform_dynamic(
                corpus, timestamps=timestamps, embeddings=embeddings, bins=bins
            )
        dtm = self.vectorizer.transform(corpus)  # type: ignore
        try:
            classes = self.classes_
        except AttributeError:
            classes = list(range(self.components_.shape[0]))
        res = TopicData(
            corpus=corpus,
            document_term_matrix=dtm,
            vocab=self.get_vocab(),
            document_topic_matrix=document_topic_matrix,
            document_representation=embeddings,
            topic_term_matrix=self.components_,  # type: ignore
            transform=getattr(self, "transform", None),
            topic_names=self.topic_names,
            classes=classes,
            temporal_components=self.temporal_components_,
            temporal_importance=self.temporal_importance_,
            time_bin_edges=self.time_bin_edges,
        )
        return res



import re
from typing import Optional

import torch
from transformers import (
    AutoModelForTokenClassification,
    AutoTokenizer,
    pipeline,
)

from turftopic.analyzers.base import Analyzer

GLINER_DESC_PROMPT = "Summarize the following:"
GLINER_NAME_PROMPT = "What is the common topic of these documents?"


class UTCAnalyzer(Analyzer):
    """Analyze topic model with a universal token classification model.

    Parameters
    ----------
    model_name: str, default 'knowledgator/UTC-DeBERTa-base-v2'
        GliNER model to use for analyses.
    use_summaries: bool, default False
        Indicates whether the language model should summarize documents before
        analyzing the topics.
    context: str, default None
        Additional context provided to the analyzer for analysis.
        e.g. "Analyze topics from blog posts related to morality and religion"
    system_prompt: str, default None
        Ignored, exists for compatibility
    summary_prompt: str, default None
        Prompt to use for abstractive summarization.
    namer_prompt: str, default None
        Prompt template for naming topics.
    description_prompt: str, default None
        Prompt template for generating topic descriptions.
    device: str, default "cpu"
        ID of the device to run the language model on.
    """

    def __init__(
        self,
        model_name: str = "knowledgator/UTC-DeBERTa-base-v2",
        context: Optional[str] = None,
        use_summaries: bool = False,
        system_prompt: Optional[str] = None,
        summary_prompt: Optional[str] = None,
        namer_prompt: Optional[str] = GLINER_NAME_PROMPT,
        description_prompt: Optional[str] = GLINER_DESC_PROMPT,
        device: str = "cpu",
        threshold: float = 0.5,
        max_tokens: int = 256,
    ):
        self.threshold = threshold
        self.device = device
        self.model_name = model_name
        self.max_tokens = max_tokens
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name)
        self.model = AutoModelForTokenClassification.from_pretrained(
            self.model_name,
        )
        self.nlp = pipeline(
            "ner",
            model=self.model,
            tokenizer=self.tokenizer,
            aggregation_strategy="first",
            device=self.device,
        )
        self.summary_prompt = summary_prompt or self.summary_prompt
        self.namer_prompt = namer_prompt or self.namer_prompt
        self.description_prompt = description_prompt or self.description_prompt
        self.use_summaries = use_summaries

    def summarize_document(self, document: str) -> str:
        prompt = self.summary_prompt
        prompt = re.sub(r"{[^}]*}*", "", prompt)
        return "\n".join(self.process(prompt, document))

    def describe_topic(
        self,
        keywords: list[str],
        documents: Optional[list] = None,
    ):
        prompt = self.description_prompt
        _keys = ", ".join(keywords)
        text = """ ## Top Keywords:
        {keywords}
        """.format(
            keywords=_keys
        )
        if documents:
            text += "## Top Documents:"
            for document in documents:
                text += f"\n - {document}"
        items = self.process(prompt, text)
        items = [item.lower().split() for item in items]
        items = [s for s in items if s]
        return "•" + "\n • ".join(items)

    def name_topic(
        self,
        keywords: list[str],
        documents: Optional[list] = None,
    ) -> str:
        prompt = self.namer_prompt
        _keys = ", ".join(keywords)
        text = """ ## Top Keywords:
        {keywords}
        """.format(
            keywords=_keys
        )
        if documents:
            text += "## Top Documents:"
            for document in documents:
                text += f"\n - {document}"
        spans = self.process(prompt, text)
        spans = [s.strip().lower() for s in spans if s.strip()]
        spans = set(spans)
        return " | ".join(spans)

    def generate_text(self, prompt: str) -> str:
        return " ".join(self.process(prompt="", text=prompt))

    def process(self, prompt: str, text: str) -> str:
        # Concatenate text and prompt for full input
        input_ = f"{prompt}\n{text}"
        model_input = next(
            self.nlp.preprocess(
                input_,
                tokenizer_params={"max_length": self.max_tokens},
                is_split_into_words=False,
                delimiter=" ",
                offset_mapping=None,
            )
        )
        with torch.no_grad():
            model_output = self.nlp._forward(model_input)
        results = self.nlp.postprocess(
            [model_output], aggregation_strategy="first"
        )
        spans = []
        prompt_length = len(prompt)  # Get prompt length
        for result in results:
            # check whether score is higher than treshold
            if result["score"] < self.threshold:
                continue
            # Adjust indices by subtracting prompt length
            start = result["start"] - prompt_length
            # If indexes belongs to the prompt - continue
            if start < 0:
                continue
            end = result["end"] - prompt_length
            # Extract span from original text using adjusted indices
            span = text[start:end]
            # Remove unnecessary white space
            span = " ".join(span.split())
            spans.append(span)
        return spans



import itertools
from typing import Iterable

from sklearn.feature_extraction.text import CountVectorizer

from turftopic.base import Encoder


def batched(iterable, n: int) -> Iterable[list[str]]:
    "Batch data into tuples of length n. The last batch may be shorter."
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := list(itertools.islice(it, n)):
        yield batch


class AttentionKeywordExtractor:
    def __init__(
        self, top_n: int, encoder: Encoder, vectorizer: CountVectorizer
    ):
        self.top_n = top_n
        self.encoder = encoder
        self.vectorizer = vectorizer
        self.key_to_index: dict[str, int] = {}
        self.term_embeddings: Optional[np.ndarray] = None

    @property
    def is_encoder_promptable(self) -> bool:
        prompts = getattr(self.encoder, "prompts", None)
        if prompts is None:
            return False
        if ("query" in prompts) and ("passage" in prompts):
            return True

    @property
    def n_vocab(self) -> int:
        return len(self.key_to_index)

    def _add_terms(self, new_terms: list[str]):
        for term in new_terms:
            self.key_to_index[term] = self.n_vocab
        if not self.is_encoder_promptable:
            term_encodings = self.encoder.encode(new_terms)
        else:
            term_encodings = self.encoder.encode(
                new_terms, prompt_name="passage"
            )
        if self.term_embeddings is not None:
            self.term_embeddings = np.concatenate(
                (self.term_embeddings, term_encodings), axis=0
            )
        else:
            self.term_embeddings = term_encodings

    def batch_extract_keywords(
        self,
        documents: list[str],
        embeddings: Optional[np.ndarray] = None,
    ) -> list[dict[str, float]]:
        if not len(documents):
            return []
        if embeddings is None:
            if not self.is_encoder_promptable:
                embeddings = self.encoder.encode(documents)
            else:
                embeddings = self.encoder.encode(
                    documents, prompt_name="query"
                )
        if len(embeddings) != len(documents):
            raise ValueError(
                "Number of documents doesn't match number of embeddings."
            )
        keywords = []
        vectorizer = clone(self.vectorizer)
        document_term_matrix = vectorizer.fit_transform(documents)
        batch_vocab = vectorizer.get_feature_names_out()
        new_terms = list(set(batch_vocab) - set(self.key_to_index.keys()))
        if len(new_terms):
            self._add_terms(new_terms)
        total = embeddings.shape[0]
        for i in range(total):
            terms = document_term_matrix[i, :].todense()
            embedding = embeddings[i].reshape(1, -1)
            mask = terms > 0
            if not np.any(mask):
                keywords.append(dict())
                continue
            important_terms = np.ravel(np.asarray(mask))
            word_embeddings = [
                self.term_embeddings[self.key_to_index[term]]
                for term in batch_vocab[important_terms]
            ]
            if self.term_embeddings.shape[1] != embeddings.shape[1]:
                raise ValueError(
                    NOT_MATCHING_ERROR.format(
                        n_dims=embeddings.shape[1],
                        n_word_dims=self.term_embeddings.shape[1],
                    )
                )
            sim = cosine_similarity(embedding, word_embeddings).astype(
                np.float64
            )
            sim = np.ravel(sim)
            kth = min(self.top_n, len(sim) - 1)
            top = np.argpartition(-sim, kth)[:kth]
            top_words = batch_vocab[important_terms][top]
            top_sims = [sim for sim in sim[top] if sim > 0]
            keywords.append(dict(zip(top_words, top_sims)))
        return keywords



import warnings
from typing import Optional

import numpy as np
import torch
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.metrics.pairwise import cosine_similarity
from tqdm import trange
from transformers import AutoTokenizer, BertModel


class ColBERTKeywordExtractor:
    def __init__(
        self,
        top_n: int,
        model_name: str,
        vectorizer: CountVectorizer,
        batch_size: int = 32,
    ):
        self.top_n = top_n
        self.model_name = model_name
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.model = BertModel.from_pretrained(model_name)
        self.vectorizer = vectorizer
        self.batch_size = batch_size
        self.key_to_index: dict[str, int] = {}
        self.term_embeddings: Optional[np.ndarray] = None

    def encode_batch(self, sentences: list[str]) -> list[np.ndarray]:
        with torch.no_grad():
            inputs = self.tokenizer(
                sentences, return_tensors="pt", padding=True
            )
            outputs = self.model(**inputs)
        hidden_state = outputs.last_hidden_state
        embeddings = []
        for h, m in zip(hidden_state, inputs["attention_mask"]):
            embeddings.append(h[m > 0].numpy())
        return embeddings

    @property
    def vocab(self) -> np.ndarray:
        res = [""] * self.n_vocab
        for key, index in self.key_to_index.items():
            res[index] = key
        return np.array(res)

    @property
    def n_vocab(self) -> int:
        return len(self.key_to_index)

    def _add_terms(self, new_terms: list[str]):
        for term in new_terms:
            self.key_to_index[term] = self.n_vocab
        term_encodings = self.encode_batch(new_terms)
        term_encodings = np.stack(
            [np.mean(_t, axis=0) for _t in term_encodings]
        )
        if self.term_embeddings is not None:
            self.term_embeddings = np.concatenate(
                (self.term_embeddings, term_encodings), axis=0
            )
        else:
            self.term_embeddings = term_encodings

    def batch_extract_keywords(
        self,
        documents: list[str],
        embeddings: Optional[np.ndarray] = None,
        seed_embedding: Optional[np.ndarray] = None,
        fitting: bool = True,
    ) -> list[dict[str, float]]:
        if not len(documents):
            return []
        if embeddings is not None:
            warnings.warn(
                "embeddings parameter specified, but get ignored when using ColBERT."
            )
        keywords = []
        if fitting:
            document_term_matrix = self.vectorizer.fit_transform(documents)
        else:
            document_term_matrix = self.vectorizer.transform(documents)
        batch_vocab = self.vectorizer.get_feature_names_out()
        new_terms = list(set(batch_vocab) - set(self.key_to_index.keys()))
        if len(new_terms):
            self._add_terms(new_terms)
        for i in trange(
            0, len(documents), self.batch_size, desc="Extracting keywords"
        ):
            _docs = documents[i : i + self.batch_size]
            _embs = self.encode_batch(_docs)
            for j in range(0, self.batch_size):
                terms = document_term_matrix[i + j, :].todense()
                mask = terms > 0
                if not np.any(mask):
                    keywords.append(dict())
                    continue
                important_terms = np.ravel(np.asarray(mask))
                word_embeddings = [
                    self.term_embeddings[self.key_to_index[term]]
                    for term in batch_vocab[important_terms]
                ]
                sim = cosine_similarity(word_embeddings, _embs[j])
                maxsim = np.max(sim, axis=1)
                kth = min(self.top_n, len(maxsim) - 1)
                top = np.argpartition(-maxsim, kth)[:kth]
                top_words = batch_vocab[important_terms][top]
                top_sims = [sim for sim in maxsim[top] if sim > 0]
                keywords.append(dict(zip(top_words, top_sims)))
        return keywords



import itertools
import warnings
from collections import defaultdict
from datetime import datetime
from functools import partial
from typing import Iterable, Literal, Optional, Union

import igraph as ig
import numpy as np
import scipy.sparse as spr
from sklearn.base import clone
from sklearn.decomposition._nmf import (
    NMF,
    MiniBatchNMF,
    _initialize_nmf,
    _update_coordinate_descent,
)
from sklearn.exceptions import NotFittedError
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.utils import check_array
from sklearn.utils.validation import check_non_negative

from turftopic.base import Encoder
from turftopic.optimization import (
    decomposition_gaussian_bic,
    optimize_n_components,
)

NOT_MATCHING_ERROR = (
    "Document embedding dimensionality ({n_dims}) doesn't match term embedding dimensionality ({n_word_dims}). "
    + "Perhaps you are using precomputed embeddings but forgot to pass an encoder to your model. "
    + "Try to initialize the model with the encoder you used for computing the embeddings."
)


def batched(iterable, n: int) -> Iterable[list[str]]:
    "Batch data into tuples of length n. The last batch may be shorter."
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := list(itertools.islice(it, n)):
        yield batch


def _match_terms(
    keywords: list[dict],
    vocab: np.ndarray,
    vocab_embeddings: np.ndarray,
    threshold: float = 0.9,
) -> list[dict]:
    """Matches similar terms into a single, cross-lingual term."""
    similarity = cosine_similarity(vocab_embeddings, vocab_embeddings)
    np.fill_diagonal(similarity, 0)
    similarity[similarity < threshold] = 0
    edges = zip(*np.nonzero(similarity > 0))
    g = ig.Graph(n=len(vocab), edges=edges)
    components = g.connected_components()
    old_to_new = {}
    for component in components:
        component_name = "-".join(vocab[component][:3])
        for i_word in component:
            old_to_new[vocab[i_word]] = component_name
    joint_keywords = []
    for entry in keywords:
        joint_entry = defaultdict(lambda: 0)
        for word, value in entry.items():
            new_word = old_to_new[word]
            joint_entry[new_word] += value
        joint_keywords.append(joint_entry)
    return joint_keywords


def fit_timeslice(
    X,
    W,
    H,
    tol=1e-4,
    max_iter=200,
    l1_reg_W=0,
    l1_reg_H=0,
    l2_reg_W=0,
    l2_reg_H=0,
    verbose=0,
    shuffle=False,
    random_state=None,
):
    """Fits topic_term_matrix based on a precomputed document_topic_matrix.
    This is used to get temporal components in dynamic KeyNMF.
    """
    Ht = check_array(H.T, order="C")
    if random_state is None:
        rng = np.random.mtrand._rand
    else:
        rng = np.random.RandomState(random_state)
    for n_iter in range(1, max_iter + 1):
        violation = 0.0
        violation += _update_coordinate_descent(
            X.T, Ht, W, l1_reg_H, l2_reg_H, shuffle, rng
        )
        if n_iter == 1:
            violation_init = violation
        if violation_init == 0:
            break
        if verbose:
            print("violation:", violation / violation_init)
        if violation / violation_init <= tol:
            if verbose:
                print("Converged at iteration", n_iter + 1)
            break
    return W, Ht.T, n_iter


class SBertKeywordExtractor:
    def __init__(
        self,
        top_n: int,
        encoder: Encoder,
        vectorizer: CountVectorizer,
        metric: Literal["cosine", "dot"],
    ):
        self.top_n = top_n
        self.encoder = encoder
        self.vectorizer = vectorizer
        self.key_to_index: dict[str, int] = {}
        self.term_embeddings: Optional[np.ndarray] = None
        self.metric = metric

    @property
    def vocab(self) -> np.ndarray:
        res = [""] * self.n_vocab
        for key, index in self.key_to_index.items():
            res[index] = key
        return np.array(res)

    def match_terms(
        self, keywords: list[dict], threshold: float = 0.9
    ) -> list[dict]:
        return _match_terms(
            keywords, self.vocab, self.term_embeddings, threshold=threshold
        )

    @property
    def is_encoder_promptable(self) -> bool:
        prompts = getattr(self.encoder, "prompts", None)
        if prompts is None:
            return False
        if ("query" in prompts) and ("passage" in prompts):
            return True

    def encode(
        self, texts: Iterable[str], prompt_name: str = None
    ) -> np.ndarray:
        if not hasattr(self.encoder, "encode"):
            return self.encoder.get_text_embeddings(list(texts))
        if (prompt_name is not None) and (self.is_encoder_promptable):
            return self.encoder.encode(texts, prompt_name=prompt_name)
        return self.encoder.encode(texts)

    @property
    def n_vocab(self) -> int:
        return len(self.key_to_index)

    def _add_terms(self, new_terms: list[str]):
        for term in new_terms:
            self.key_to_index[term] = self.n_vocab
        term_encodings = self.encode(new_terms, prompt_name="passage")
        if self.term_embeddings is not None:
            self.term_embeddings = np.concatenate(
                (self.term_embeddings, term_encodings), axis=0
            )
        else:
            self.term_embeddings = term_encodings

    def batch_extract_keywords(
        self,
        documents: list[str],
        embeddings: Optional[np.ndarray] = None,
        seed_embedding: Optional[np.ndarray] = None,
        seed_exponent: float = 4.0,
        fitting: bool = True,
    ) -> list[dict[str, float]]:
        if not len(documents):
            return []
        if embeddings is None:
            embeddings = self.encode(documents, prompt_name="query")
        if len(embeddings) != len(documents):
            raise ValueError(
                "Number of documents doesn't match number of embeddings."
            )
        keywords = []
        if fitting:
            document_term_matrix = self.vectorizer.fit_transform(documents)
        else:
            document_term_matrix = self.vectorizer.transform(documents)
        batch_vocab = self.vectorizer.get_feature_names_out()
        new_terms = list(set(batch_vocab) - set(self.key_to_index.keys()))
        if len(new_terms):
            self._add_terms(new_terms)
        total = embeddings.shape[0]
        # Relevance based on similarity to seed embedding
        document_relevance = np.ones(embeddings.shape[0])
        if seed_embedding is not None:
            if self.metric == "cosine":
                document_relevance = cosine_similarity(
                    [seed_embedding], embeddings
                )[0]
            else:
                document_relevance = np.dot(embeddings, seed_embedding)
            document_relevance[document_relevance < 0] = 0
            document_relevance = np.power(document_relevance, seed_exponent)
        for i in range(total):
            terms = document_term_matrix[i, :].todense()
            embedding = embeddings[i].reshape(1, -1)
            mask = terms > 0
            if not np.any(mask):
                keywords.append(dict())
                continue
            important_terms = np.ravel(np.asarray(mask))
            word_embeddings = [
                self.term_embeddings[self.key_to_index[term]]
                for term in batch_vocab[important_terms]
            ]
            if self.term_embeddings.shape[1] != embeddings.shape[1]:
                raise ValueError(
                    NOT_MATCHING_ERROR.format(
                        n_dims=embeddings.shape[1],
                        n_word_dims=self.term_embeddings.shape[1],
                    )
                )
            if self.metric == "cosine":
                sim = cosine_similarity(embedding, word_embeddings)
                sim = np.ravel(sim)
            else:
                sim = np.dot(word_embeddings, embedding[0]).T
            # If a seed is specified, we multiply by the document's relevance
            sim = document_relevance[i] * sim
            kth = min(self.top_n, len(sim) - 1)
            top = np.argpartition(-sim, kth)[:kth]
            top_words = batch_vocab[important_terms][top]
            top_sims = [sim for sim in sim[top] if sim > 0]
            keywords.append(dict(zip(top_words, top_sims)))
        return keywords


class KeywordNMF:
    def __init__(
        self,
        n_components: Union[int, Literal["auto"]],
        seed: Optional[int] = None,
        top_n: Optional[int] = None,
    ):
        self.n_components = n_components
        self.key_to_index: dict[str, int] = {}
        self.index_to_key: list[str] = []
        self.top_n = top_n
        # n_components * n_vocab
        self.components: Optional[np.ndarray] = None
        self.seed = seed
        self.temporal_components: Optional[np.ndarray] = None
        self.temporal_importance_: Optional[np.ndarray] = None

    def prune_keywords(self, keywords: dict[str, float]) -> dict[str, float]:
        """If there are more keywords than allowed, this prunes them."""
        if (self.top_n is None) or (self.top_n >= len(keywords)):
            return keywords
        words, similarities = zip(*keywords.items())
        similarities = np.array(similarities)
        selected = np.argsort(-similarities)[: self.top_n]
        items = [(words[i], similarities[i]) for i in selected]
        return dict(items)

    @property
    def n_vocab(self) -> int:
        return len(self.index_to_key)

    def _add_word_components(self, X: spr.csr_matrix):
        """Initializes components for novel vocabulary."""
        _, H = _initialize_nmf(X, self.n_components, random_state=self.seed)
        if self.components is None:
            self.components = H
        else:
            n_new = X.shape[1] - self.components.shape[1]
            if n_new:
                self.components = np.concatenate(
                    (self.components, H[:, -n_new:]), axis=1
                )
        if self.temporal_components is not None:
            n_new = X.shape[1] - self.temporal_components.shape[-1]
            if n_new:
                new_comps = H[:, -n_new:]
                new_comps = np.broadcast_to(
                    new_comps,
                    (self.temporal_components.shape[0], *new_comps.shape),
                )
                self.temporal_components = np.concatenate(
                    (self.temporal_components, new_comps), axis=-1
                )

    def vectorize(
        self, keywords: list[dict[str, float]], fitting: bool = False
    ) -> spr.csr_array:
        indices = []
        indptr = [0]
        values = []
        for k in keywords:
            k = self.prune_keywords(k)
            for w, v in k.items():
                # Adding vocab item if missing
                if (w not in self.key_to_index) and fitting:
                    self.key_to_index[w] = self.n_vocab
                    self.index_to_key.append(w)
                if w in self.key_to_index:
                    indices.append(self.key_to_index[w])
                    values.append(v)
            indptr.append(len(indices))
        shape = (len(indptr) - 1, self.n_vocab)
        document_term_matrix = spr.csr_matrix(
            (values, indices, indptr), shape=shape
        )
        return document_term_matrix

    def fit_transform(self, keywords: list[dict[str, float]]) -> np.ndarray:
        X = self.vectorize(keywords, fitting=True)
        if self.n_components == "auto":
            # Finding N components with BIC
            bic_fn = partial(
                decomposition_gaussian_bic,
                decomp_class=NMF,
                X=X,
            )
            n_components = optimize_n_components(bic_fn, min_n=1, verbose=True)
            self.n_components = n_components
        check_non_negative(X, "NMF (input X)")
        W, H = _initialize_nmf(X, self.n_components, random_state=self.seed)
        W, H, self.n_iter = NMF(
            self.n_components, init="custom", random_state=self.seed
        )._fit_transform(X, W=W, H=H, update_H=True)
        self.components = H.astype(X.dtype)
        return W

    def transform(self, keywords: list[dict[str, float]]):
        if self.components is None:
            raise NotFittedError(
                "Can't transform() if the model has not been fitted."
            )
        X = self.vectorize(keywords, fitting=False)
        check_non_negative(X, "NMF (input X)")
        W, _, _ = NMF(
            self.n_components, init="custom", random_state=self.seed
        )._fit_transform(X, W=None, H=self.components, update_H=False)
        return W.astype(X.dtype)

    def partial_fit(self, keyword_batch: list[dict[str, float]]):
        if self.n_components == "auto":
            raise ValueError(
                "Cannot infer number of components with BIC when online fitting the model."
            )
        X = self.vectorize(keyword_batch, fitting=True)
        try:
            check_non_negative(X, "NMF (input X)")
            self._add_word_components(X)
            W, _ = _initialize_nmf(
                X, self.n_components, random_state=self.seed
            )
            _minibatchnmf = MiniBatchNMF(
                self.n_components, init="custom", random_state=self.seed
            ).partial_fit(X, W=W, H=self.components)
            self.components = _minibatchnmf.components_.astype(X.dtype)
        except ValueError as e:
            warnings.warn(f"Batch failed with error: {e}, skipping.")
            return self
        return self

    def fit_transform_dynamic(
        self,
        keywords: list[dict[str, float]],
        time_labels: np.ndarray,
        time_bin_edges: list[datetime],
    ) -> np.ndarray:
        self.time_bin_edges = time_bin_edges
        n_bins = len(time_bin_edges) - 1
        document_term_matrix = self.vectorize(keywords, fitting=True)
        check_non_negative(document_term_matrix, "NMF (input X)")
        if self.n_components == "auto":
            # Finding N components with BIC
            bic_fn = partial(
                decomposition_gaussian_bic,
                decomp_class=NMF,
                X=X,
            )
            n_components = optimize_n_components(bic_fn, verbose=True)
            self.n_components = n_components
        document_topic_matrix, H = _initialize_nmf(
            document_term_matrix,
            self.n_components,
            random_state=self.seed,
        )
        document_topic_matrix, H, self.n_iter = NMF(
            self.n_components, init="custom", random_state=self.seed
        )._fit_transform(
            document_term_matrix, W=document_topic_matrix, H=H, update_H=True
        )
        self.components = H.astype(document_term_matrix.dtype)
        n_comp, n_vocab = self.components.shape
        self.temporal_components = np.zeros(
            (n_bins, n_comp, n_vocab), dtype=document_term_matrix.dtype
        )
        self.temporal_importance_ = np.zeros((n_bins, n_comp))
        for label in np.unique(time_labels):
            idx = np.nonzero(time_labels == label)
            X = document_term_matrix[idx]
            W = document_topic_matrix[idx]
            _, H = _initialize_nmf(
                X, self.components.shape[0], random_state=self.seed
            )
            _, H, _ = fit_timeslice(X, W, H, random_state=self.seed)
            self.temporal_components[label] = H
            topic_importances = np.squeeze(np.asarray(W.sum(axis=0)))
            self.temporal_importance_[label] = topic_importances
        return document_topic_matrix

    def partial_fit_dynamic(
        self,
        keyword_batch: list[dict[str, float]],
        time_labels: np.ndarray,
        time_bin_edges: list[datetime],
    ) -> np.ndarray:
        if self.temporal_components is None:
            self.fit_transform_dynamic(
                keyword_batch, time_labels, time_bin_edges
            )
        else:
            document_term_matrix = self.vectorize(keyword_batch, fitting=True)
            check_non_negative(document_term_matrix, "NMF (input X)")
            self._add_word_components(document_term_matrix)
            document_topic_matrix = self.transform(keyword_batch)
            _minibatchnmf = MiniBatchNMF(
                self.n_components, init="custom", random_state=self.seed
            ).partial_fit(
                document_term_matrix,
                W=document_topic_matrix,
                H=self.components,
            )
            self.components = _minibatchnmf.components_.astype(
                document_term_matrix.dtype
            )
            document_topic_matrix = self.transform(keyword_batch)
            for label in np.unique(time_labels):
                idx = np.nonzero(time_labels == label)
                X = document_term_matrix[idx]
                W = document_topic_matrix[idx]
                _minibatchnmf = MiniBatchNMF(
                    self.n_components, init="custom", random_state=self.seed
                ).partial_fit(
                    X,
                    W=W,
                    H=self.temporal_components[label],
                )
                self.temporal_components[label] = _minibatchnmf.components_
                topic_importances = np.squeeze(np.asarray(W.sum(axis=0)))
                self.temporal_importance_[label] += topic_importances



import itertools
import re

import numpy as np
from sklearn.feature_extraction.text import CountVectorizer

try:
    import jieba
except ModuleNotFoundError as e:
    raise ModuleNotFoundError(
        "Jieba is not installed on your machine, please run `pip install turftopic[jieba]` before using Chinese models in Turftopic."
    ) from e


def tokenize_zh(text: str):
    # Remove nonalphanumeric characters
    text = re.sub(r"\W*", "", text)
    # Tokenize text
    tokens = jieba.lcut(text)
    # Remove empty tokens
    tokens = [token for token in tokens if token.strip()]
    return tokens


chinese_stop_words = """
一
一個
一些
一何
一切
一則
一方面
一旦
一來
一樣
一種
一般
一轉眼
七
萬一
三
上
上下
下
不
不僅
不但
不光
不單
不只
不外乎
不如
不妨
不盡
不盡然
不得
不怕
不惟
不成
不拘
不料
不是
不比
不然
不特
不獨
不管
不至於
不若
不論
不過
不問
與
與其
與其說
與否
與此同時
且
且不說
且說
兩者
個
個別
中
臨
為
為了
為什麼
為何
為止
為此
為著
乃
乃至
乃至於
麼
之
之一
之所以
之類
烏乎
乎
乘
九
也
也好
也罷
了
二
二來
於
於是
於是乎
雲雲
雲爾
五
些
亦
人
人們
人家
什
什麼
什麼樣
今
介於
仍
仍舊
從
從此
從而
他
他人
他們
他們們
以
以上
以為
以便
以免
以及
以故
以期
以來
以至
以至於
以致
們
任
任何
任憑
會
似的
但
但凡
但是
何
何以
何況
何處
何時
余外
作為
你
你們
使
使得
例如
依
依據
依照
便於
俺
俺們
倘
倘使
倘或
倘然
倘若
借
借儻然
假使
假如
假若
做
像
兒
先不先
光
光是
全體
全部
八
六
兮
共
關於
關於具體地說
其
其一
其中
其二
其他
其餘
其它
其次
具體地說
具體說來
兼之
內
再
再其次
再則
再有
再者
再者說
再說
冒
衝
況且
幾
幾時
凡
凡是
憑
憑借
出於
出來
分
分別
則
則甚
別
別人
別處
別是
別的
別管
別說
到
前後
前此
前者
加之
加以
區
即
即令
即使
即便
即如
即或
即若
卻
去
又
又及
及
及其
及至
反之
反而
反過來
反過來說
受到
另
另一方面
另外
另悉
只
只當
只怕
只是
只有
只消
只要
只限
叫
叮咚
可
可以
可是
可見
各
各個
各位
各種
各自
同
同時
後
後者
向
向使
向著
嚇
嗎
否則
吧
吧噠
含
吱
呀
呃
嘔
唄
嗚
嗚呼
呢
呵
呵呵
呸
呼哧
咋
和
咚
咦
咧
咱
咱們
咳
哇
哈
哈哈
哉
哎
哎呀
哎喲
嘩
喲
哦
哩
哪
哪個
哪些
哪兒
哪天
哪年
哪怕
哪樣
哪邊
哪裡
哼
哼唷
唉
唯有
啊
啐
啥
啦
啪達
啷當
餵
喏
喔唷
嘍
嗡
嗡嗡
嗬
嗯
噯
嘎
嘎登
噓
嘛
嘻
嘿
嘿嘿
四
因
因為
因了
因此
因著
因而
固然
在
在下
在於
地
基於
處在
多
多麼
多少
大
大家
她
她們
好
如
如上
如上所述
如下
如何
如其
如同
如是
如果
如此
如若
始而
孰料
孰知
寧
寧可
寧願
寧肯
它
它們
對
對於
對待
對方
對比
將
小
爾
爾後
爾爾
尚且
就
就是
就是了
就是說
就算
就要
盡
儘管
儘管如此
豈但
己
已
已矣
巴
巴巴
年
並
並且
庶乎
庶幾
開外
開始
歸
歸齊
當
當地
當然
當著
彼
彼時
彼此
往
待
很
得
得了
怎
怎麼
怎麼辦
怎麼樣
怎奈
怎樣
總之
總的來看
總的來說
總的說來
總而言之
恰恰相反
您
惟其
慢說
我
我們
或
或則
或是
或曰
或者
截至
所
所以
所在
所幸
所有
才
才能
打
打從
把
抑或
拿
按
按照
換句話說
換言之
據
據此
接著
故
故此
故而
旁人
無
無寧
無論
既
既往
既是
既然
日
時
時候
是
是以
是的
更
曾
替
替代
最
月
有
有些
有關
有及
有時
有的
望
朝
朝著
本
本人
本地
本著
本身
來
來著
來自
來說
極了
果然
果真
某
某個
某些
某某
根據
歟
正值
正如
正巧
正是
此
此地
此處
此外
此時
此次
此間
毋寧
每
每當
比
比及
比如
比方
沒奈何
沿
沿著
漫說
點
焉
然則
然後
然而
照
照著
猶且
猶自
甚且
甚麼
甚或
甚而
甚至
甚至於
用
用來
由
由於
由是
由此
由此可見
的
的確
的話
直到
相對而言
省得
看
眨眼
著
著呢
矣
矣乎
矣哉
離
秒
稱
竟而
第
等
等到
等等
簡言之
管
類如
緊接著
縱
縱令
縱使
縱然
經
經過
結果
給
繼之
繼後
繼而
綜上所述
罷了
者
而
而且
而況
而後
而外
而已
而是
而言
能
能否
騰
自
自個兒
自從
自各兒
自後
自家
自己
自打
自身
至
至於
至今
至若
致
般的
若
若夫
若是
若果
若非
莫不然
莫如
莫若
雖
雖則
雖然
雖說
被
要
要不
要不是
要不然
要麼
要是
譬喻
譬如
讓
許多
論
設使
設或
設若
誠如
誠然
該
說
說來
請
諸
諸位
諸如
誰
誰人
誰料
誰知
賊死
賴以
趕
起
起見
趁
趁著
越是
距
跟
較
較之
邊
過
還
還是
還有
還要
這
這一來
這個
這麼
這麼些
這麼樣
這麼點兒
這些
這會兒
這兒
這就是說
這時
這樣
這次
這般
這邊
這裡
進而
連
連同
逐步
通過
遵循
遵照
那
那個
那麼
那麼些
那麼樣
那些
那會兒
那兒
那時
那樣
那般
那邊
那裡
都
鄙人
鑒於
針對
阿
除
除了
除外
除開
除此之外
除非
隨
隨後
隨時
隨著
難道說
零
非
非但
非徒
非特
非獨
靠
順
順著
首先
︿
！
＃
＄
％
＆
（
）
＊
＋
，
０
１
２
３
４
５
６
７
８
９
：
；
＜
＞
？
＠
［
］
｛
｜
｝
～
￥
i
me
my
myself
we
our
ours
ourselves
you
you're
you've
you'll
you'd
your
yours
yourself
yourselves
he
him
his
himself
she
she's
her
hers
herself
it
it's
its
itself
they
them
their
theirs
themselves
what
which
who
whom
this
that
that'll
these
those
am
is
are
was
were
be
been
being
have
has
had
having
do
does
did
doing
a
an
the
and
but
if
or
because
as
until
while
of
at
by
for
with
about
against
between
into
through
during
before
after
above
below
to
from
up
down
in
out
on
off
over
under
again
further
then
once
here
there
when
where
why
how
all
any
both
each
few
more
most
other
some
such
no
nor
not
only
own
same
so
than
too
very
s
t
can
will
just
don
don't
should
should've
now
d
ll
m
o
re
ve
y
ain
aren
aren't
couldn
couldn't
didn
didn't
doesn
doesn't
hadn
hadn't
hasn
hasn't
haven
haven't
isn
isn't
ma
mightn
mightn't
mustn
mustn't
needn
needn't
shan
shan't
shouldn
shouldn't
wasn
wasn't
weren
weren't
won
won't
wouldn
wouldn't
的
了
和
是
就
都
而
及
與
著
或
一個
沒有
我們
你們
妳們
他們
她們
是否
。
,
「
」
、
‧
《
》
〈
〉
——
—
～
【
】
［
］
（
）
：
；
？
！
︿
！
＃
＄
％
＆
（
）
＊
＋
，
０
１
２
３
４
５
６
７
８
９
：
；
＜
＞
？
＠
［
］
｛
｜
｝
～
￥
(
)
""".split()

# Pretokenizing the stop word list
chinese_stop_words = list(
    itertools.chain.from_iterable(map(tokenize_zh, chinese_stop_words))
)


class ChineseCountVectorizer(CountVectorizer):
    """Chinese count vectorizer. Does word segmentation with Jieba, and includes a chinese stop words list.
    You have to specify stop_words="chinese" for this to kick into effect.
    """

    def __init__(
        self,
        *,
        input="content",
        encoding="utf-8",
        decode_error="strict",
        strip_accents=None,
        lowercase=True,
        preprocessor=None,
        tokenizer=tokenize_zh,
        stop_words=None,
        token_pattern=r"(?u)\b\w\w+\b",
        ngram_range=(1, 1),
        analyzer="word",
        max_df=1.0,
        min_df=1,
        max_features=None,
        vocabulary=None,
        binary=False,
        dtype=np.int64,
    ):
        if stop_words == "chinese":
            stop_words = chinese_stop_words
        super().__init__(
            input=input,
            encoding=encoding,
            decode_error=decode_error,
            strip_accents=strip_accents,
            lowercase=lowercase,
            preprocessor=preprocessor,
            tokenizer=tokenizer,
            stop_words=stop_words,
            token_pattern=token_pattern,
            ngram_range=ngram_range,
            analyzer=analyzer,
            max_df=max_df,
            min_df=min_df,
            max_features=max_features,
            vocabulary=vocabulary,
            binary=binary,
            dtype=dtype,
        )


def default_chinese_vectorizer():
    return ChineseCountVectorizer(stop_words="chinese", min_df=10)



from typing import Optional

from transformers import AutoModelForCausalLM, AutoTokenizer, pipeline

from turftopic.analyzers.base import Analyzer


def remove_thinking_trace(response: str) -> str:
    """Removes think tag and all tokens in-between"""
    if "<think>" not in response:
        return response
    before, after = response.split("<think>")
    between, after = after.split("</think>")
    return before + after


class LLMAnalyzer(Analyzer):
    """Analyze topic model with an open LLM from HF Hub.

    Parameters
    ----------
    model_name: str, default 'HuggingFaceTB/SmolLM3-3B'
        Open LLM to use from HF Hub.
    use_summaries: bool, default False
        Indicates whether the language model should summarize documents before
        analyzing the topics.
    context: str, default None
        Additional context provided to the analyzer for analysis.
        e.g. "Analyze topics from blog posts related to morality and religion"
    system_prompt: str, default None
        Ignored, exists for compatibility
    summary_prompt: str, default None
        Prompt to use for abstractive summarization.
    namer_prompt: str, default None
        Prompt template for naming topics.
    description_prompt: str, default None
        Prompt template for generating topic descriptions.
    device: str, default "cpu"
        ID of the device to run the language model on.
    max_new_tokens: int, default 32768
        Max new tokens to generate when analyzing.
    enable_thinking: bool, default False
        Indicates whether thinking mode should be enabled.
    """

    def __init__(
        self,
        model_name: str = "HuggingFaceTB/SmolLM3-3B",
        context: Optional[str] = None,
        use_summaries: bool = False,
        system_prompt: Optional[str] = None,
        summary_prompt: Optional[str] = None,
        namer_prompt: Optional[str] = None,
        description_prompt: Optional[str] = None,
        max_new_tokens: int = 32768,
        device: str = "cpu",
        enable_thinking: bool = False,
    ):
        self.device = device
        self.model_name = model_name
        # load the tokenizer and the model
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name)
        self.model = AutoModelForCausalLM.from_pretrained(
            self.model_name,
        ).to(self.device)
        self.summary_prompt = summary_prompt or self.summary_prompt
        self.namer_prompt = namer_prompt or self.namer_prompt
        self.description_prompt = description_prompt or self.description_prompt
        self.use_summaries = use_summaries
        self.max_new_tokens = max_new_tokens
        self.enable_thinking = enable_thinking

    def generate_text(self, prompt: str) -> str:
        thinking = "/think" if self.enable_thinking else "/no_think"
        system_prompt = self.system_prompt + thinking
        messages = [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": prompt},
        ]
        text = self.tokenizer.apply_chat_template(
            messages,
            tokenize=False,
            add_generation_prompt=True,
        )
        model_inputs = self.tokenizer([text], return_tensors="pt").to(
            self.model.device
        )
        # Generate the output
        generated_ids = self.model.generate(
            **model_inputs, max_new_tokens=32768
        )
        # Get and decode the output
        output_ids = generated_ids[0][len(model_inputs.input_ids[0]) :]
        result = self.tokenizer.decode(output_ids, skip_special_tokens=True)
        result = remove_thinking_trace(result)
        return result



import re
from typing import Union

import numpy as np
from sklearn.feature_extraction.text import CountVectorizer

try:
    import spacy
    from spacy.language import Language
except ModuleNotFoundError as e:
    raise ModuleNotFoundError(
        "SpaCy is not installed on your machine, please run `pip install turftopic[spacy]` before using SpaCy tokenizers."
    ) from e


class NounPhraseCountVectorizer(CountVectorizer):
    """Extracts Noun phrases from text using SpaCy.

    Parameters
    ----------
    nlp: spacy.Language or str, default "en_core_web_sm"
        A Spacy pipeline or its name.
    """

    def __init__(
        self,
        nlp: Union[Language, str] = "en_core_web_sm",
        *,
        input="content",
        encoding="utf-8",
        decode_error="strict",
        strip_accents=None,
        lowercase=True,
        preprocessor=None,
        tokenizer=None,
        stop_words=None,
        token_pattern=r"(?u)\b\w\w+\b",
        ngram_range=(1, 1),
        analyzer="word",
        max_df=1.0,
        min_df=1,
        max_features=None,
        vocabulary=None,
        binary=False,
        dtype=np.int64,
    ):
        self.nlp = nlp
        if isinstance(nlp, str):
            self._nlp = spacy.load(nlp)
        else:
            self._nlp = nlp
        super().__init__(
            input=input,
            encoding=encoding,
            decode_error=decode_error,
            strip_accents=strip_accents,
            lowercase=lowercase,
            preprocessor=preprocessor,
            tokenizer=tokenizer,
            stop_words=stop_words,
            token_pattern=token_pattern,
            ngram_range=ngram_range,
            analyzer=analyzer,
            max_df=max_df,
            min_df=min_df,
            max_features=max_features,
            vocabulary=vocabulary,
            binary=binary,
            dtype=dtype,
        )

    def nounphrase_tokenize(self, text: str) -> list[str]:
        doc = self._nlp(text)
        tokens = []
        for chunk in doc.noun_chunks:
            if chunk[0].is_stop:
                chunk = chunk[1:]
            phrase = chunk.text
            phrase = re.sub(r"[^\w\s]", " ", phrase)
            phrase = " ".join(phrase.split()).strip()
            if phrase:
                tokens.append(phrase)
        return tokens

    def build_tokenizer(self):
        return self.nounphrase_tokenize


class LemmaCountVectorizer(CountVectorizer):
    """Extracts lemmata from text using SpaCy.

    Parameters
    ----------
    nlp: spacy.Language or str, default "en_core_web_sm"
        A Spacy pipeline or its name.
    """

    def __init__(
        self,
        nlp: Union[Language, str] = "en_core_web_sm",
        *,
        input="content",
        encoding="utf-8",
        decode_error="strict",
        strip_accents=None,
        lowercase=True,
        preprocessor=None,
        tokenizer=None,
        stop_words=None,
        token_pattern=r"(?u)\b\w\w+\b",
        ngram_range=(1, 1),
        analyzer="word",
        max_df=1.0,
        min_df=1,
        max_features=None,
        vocabulary=None,
        binary=False,
        dtype=np.int64,
    ):
        self.nlp = nlp
        if isinstance(nlp, str):
            self._nlp = spacy.load(nlp)
        else:
            self._nlp = nlp
        super().__init__(
            input=input,
            encoding=encoding,
            decode_error=decode_error,
            strip_accents=strip_accents,
            lowercase=lowercase,
            preprocessor=preprocessor,
            tokenizer=tokenizer,
            stop_words=stop_words,
            token_pattern=token_pattern,
            ngram_range=ngram_range,
            analyzer=analyzer,
            max_df=max_df,
            min_df=min_df,
            max_features=max_features,
            vocabulary=vocabulary,
            binary=binary,
            dtype=dtype,
        )

    def lemma_tokenize(self, text: str) -> list[str]:
        doc = self._nlp(text)
        tokens = []
        for token in doc:
            if token.is_stop or not token.is_alpha:
                continue
            tokens.append(token.lemma_.strip())
        return tokens

    def build_tokenizer(self):
        return self.lemma_tokenize


class TokenCountVectorizer(CountVectorizer):
    """Tokenizes text with SpaCy using its language-specific tokenization rules and stop-word lists

    Parameters
    ----------
    language_code: str, default "en"
        Language code for the language you intend to use.
    remove_stop_words: bool, default True
        Indicates whether stop words should be removed.
    remove_nonalpha: bool, default True
        Indicates whether only tokens containing alphabetical characters should be kept.
    """

    def __init__(
        self,
        language_code: str = "en",
        remove_stop_words: bool = True,
        remove_nonalpha: bool = True,
        *,
        input="content",
        encoding="utf-8",
        decode_error="strict",
        strip_accents=None,
        lowercase=True,
        preprocessor=None,
        tokenizer=None,
        stop_words=None,
        token_pattern=r"(?u)\b\w\w+\b",
        ngram_range=(1, 1),
        analyzer="word",
        max_df=1.0,
        min_df=1,
        max_features=None,
        vocabulary=None,
        binary=False,
        dtype=np.int64,
    ):
        self.language_code = language_code
        self.remove_stop_words = remove_stop_words
        self.remove_nonalpha = remove_nonalpha
        super().__init__(
            input=input,
            encoding=encoding,
            decode_error=decode_error,
            strip_accents=strip_accents,
            lowercase=lowercase,
            preprocessor=preprocessor,
            tokenizer=tokenizer,
            stop_words=stop_words,
            token_pattern=token_pattern,
            ngram_range=ngram_range,
            analyzer=analyzer,
            max_df=max_df,
            min_df=min_df,
            max_features=max_features,
            vocabulary=vocabulary,
            binary=binary,
            dtype=dtype,
        )

    def build_tokenizer(self):
        nlp = spacy.blank(self.language_code)

        def tokenize(text: str) -> list[str]:
            doc = nlp(text)
            result = []
            for tok in doc:
                if self.remove_stop_words and tok.is_stop:
                    continue
                if self.remove_nonalpha and not tok.is_alpha:
                    continue
                result.append(tok.orth_)
            return result

        return tokenize
